#!/usr/bin/env python
# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Find contours examples

.. note:: This module has an optional dependency with sci-kit image library.
    You might need to install it if you don't already have it.
"""

import logging
import sys
import numpy
import time

logging.basicConfig()
_logger = logging.getLogger("find_contours")

from silx.gui import qt
import silx.gui.plot
from silx.gui.colors import Colormap
import silx.image.bilinear


try:
    import skimage
except ImportError:
    _logger.debug("Error while importing skimage", exc_info=True)
    skimage = None

if skimage is not None:
    try:
        from silx.image.marchingsquares._skimage import MarchingSquaresSciKitImage
    except ImportError:
        _logger.debug("Error while importing MarchingSquaresSciKitImage", exc_info=True)
        MarchingSquaresSciKitImage = None
else:
    MarchingSquaresSciKitImage = None


def rescale_image(image, shape):
    y, x = numpy.ogrid[:shape[0], :shape[1]]
    y, x = y * 1.0 * (image.shape[0] - 1) / (shape[0] - 1), x * 1.0 * (image.shape[1] - 1) / (shape[1] - 1)
    b = silx.image.bilinear.BilinearImage(image)
    # TODO: could be optimized using strides
    x2d = numpy.zeros_like(y) + x
    y2d = numpy.zeros_like(x) + y
    result = b.map_coordinates((y2d, x2d))
    return result


def create_spiral(size, nb=1, freq=100):
    half = size // 2
    y, x = numpy.ogrid[-half:half, -half:half]
    coef = 1.0 / half
    y, x = y * coef, x * coef + 0.0001
    distance = numpy.sqrt(x * x + y * y)
    angle = numpy.arctan(y / x)
    data = numpy.sin(angle * nb * 2 + distance * freq * half / 100, dtype=numpy.float32)
    return data


def create_magnetic_field(size, x1=0.0, y1=0.0, x2=0.0, y2=0.0):
    half = size // 2
    yy, xx = numpy.ogrid[-half:half, -half:half]
    coef = 1.0 / half
    yy1, xx1 = (yy + half * y1) * coef, (xx + half * x1) * coef
    distance1 = numpy.sqrt(xx1 * xx1 + yy1 * yy1)
    yy2, xx2 = (yy + half * y2) * coef, (xx + half * x2) * coef
    distance2 = numpy.sqrt(xx2 * xx2 + yy2 * yy2)
    return (numpy.arctan2(distance1, distance2) - numpy.pi * 0.25) * 1000


def create_gravity_field(size, objects):
    half = size // 2
    yy, xx = numpy.ogrid[-half:half, -half:half]
    coef = 1.0 / half

    def distance(x, y):
        yy1, xx1 = (yy + half * y) * coef, (xx + half * x) * coef
        return numpy.sqrt(xx1 ** 2 + yy1 ** 2)
    result = numpy.zeros((size, size), dtype=numpy.float32)
    for x, y, m in objects:
        result += m / distance(x, y)
    return numpy.log(result) * 1000


def create_gradient(size, dx=0, dy=0, sx=1.0, sy=1.0):
    half = size // 2
    yy, xx = numpy.ogrid[-half:half, -half:half]
    coef = 1.0 / half
    yy, xx = (yy - (dy * half)) * coef, (xx - (dx * half)) * coef + 0.0001
    distance = numpy.sqrt(xx * xx * sx + yy * yy * sy)
    return distance


def create_composite_gradient(size, dx=0, dy=0, sx=1.0, sy=1.0):
    hole = (size - 4) // 4
    gap = 10
    base = create_gradient(size + hole + gap * 4, dx, dy, sx, sy)
    result = numpy.zeros((size, size))
    width = (size - 2) // 2
    half_hole = hole // 2

    def copy_module(x1, y1, x2, y2, width, height):
        result[y1:y1 + height, x1:x1 + width] = base[y2:y2 + height, x2:x2 + width]

    y1 = 0
    y2 = 0
    copy_module(0, y1, half_hole, y2, width, hole)
    copy_module(width + 1, y1, half_hole + width, y2, width, hole)

    y1 += hole + 1
    y2 += hole + gap
    copy_module(0, y1, 0, y2, width, hole)
    copy_module(width + 1, y1, width + hole, y2, width, hole)

    y1 += hole + 1
    y2 += hole + gap
    copy_module(0, y1, half_hole, y2, width, hole)
    copy_module(width + 1, y1, half_hole + width, y2, width, hole)

    y1 += hole + 1
    y2 += hole + gap
    copy_module(0, y1, half_hole, y2, width, hole)
    copy_module(width + 1, y1, half_hole + width, y2, width, hole)

    return result


def create_value_noise(shape, octaves=8, weights=None, first_array=None):
    data = numpy.zeros(shape, dtype=numpy.float32)
    t = 2
    for i in range(octaves):
        if t > shape[0] and t > shape[1]:
            break
        if i == 0 and first_array is not None:
            d = first_array
        else:
            if weights is None:
                w = (256 >> i) - 1
            else:
                w = weights[i]
            d = numpy.random.randint(w, size=(t, t)).astype(dtype=numpy.uint8)
        d = rescale_image(d, shape)
        data = data + d
        t = t << 1
    return data


def create_island(shape, summit, under_water):
    # Force a centric shape
    first_array = numpy.zeros((4, 4), dtype=numpy.uint8)
    first_array[1:3, 1:3] = 255
    weights = [255] + [(256 >> (i)) - 1 for i in range(8)]
    data = create_value_noise(shape, octaves=7, first_array=first_array, weights=weights)
    # more slops
    data *= data
    # normalize the height
    data -= data.min()
    data = data * ((summit + under_water) / data.max()) - under_water
    return data


def createRgbaMaskImage(mask, color):
    """Generate an RGBA image where a custom color is apply to the location of
    the mask. Non masked part of the image is transparent."""
    image = numpy.zeros((mask.shape[0], mask.shape[1], 4), dtype=numpy.uint8)
    color = numpy.array(color)
    image[mask == True] = color
    return image


class FindContours(qt.QMainWindow):
    """
    This window show an example of use of a Hdf5TreeView.

    The tree is initialized with a list of filenames. A panel allow to play
    with internal property configuration of the widget, and a text screen
    allow to display events.
    """

    def __init__(self, filenames=None):
        """
        :param files_: List of HDF5 or Spec files (pathes or
            :class:`silx.io.spech5.SpecH5` or :class:`h5py.File`
            instances)
        """
        qt.QMainWindow.__init__(self)
        self.setWindowTitle("Silx HDF5 widget example")

        self.__plot = silx.gui.plot.Plot2D(parent=self)
        dummy = numpy.array([[0]])
        self.__plot.addImage(dummy, legend="image", z=-10, replace=False)
        dummy = numpy.array([[[0, 0, 0, 0]]])
        self.__plot.addImage(dummy, legend="iso-pixels", z=0, replace=False)

        self.__algo = None
        self.__polygons = []
        self.__customPolygons = []
        self.__image = None
        self.__mask = None
        self.__customValue = None

        mainPanel = qt.QWidget(self)
        layout = qt.QHBoxLayout()
        layout.addWidget(self.__createConfigurationPanel(self))
        layout.addWidget(self.__plot)
        mainPanel.setLayout(layout)

        self.setCentralWidget(mainPanel)

    def __createConfigurationPanel(self, parent):
        panel = qt.QWidget(parent=parent)
        layout = qt.QVBoxLayout()
        panel.setLayout(layout)

        self.__kind = qt.QButtonGroup(self)
        self.__kind.setExclusive(True)

        group = qt.QGroupBox(self)
        group.setTitle("Image")
        layout.addWidget(group)
        groupLayout = qt.QVBoxLayout(group)

        button = qt.QRadioButton(parent=panel)
        button.setText("Island")
        button.clicked.connect(self.generateIsland)
        button.setCheckable(True)
        button.setChecked(True)
        groupLayout.addWidget(button)
        self.__kind.addButton(button)

        button = qt.QRadioButton(parent=panel)
        button.setText("Gravity")
        button.clicked.connect(self.generateGravityField)
        button.setCheckable(True)
        groupLayout.addWidget(button)
        self.__kind.addButton(button)

        button = qt.QRadioButton(parent=panel)
        button.setText("Magnetic")
        button.clicked.connect(self.generateMagneticField)
        button.setCheckable(True)
        groupLayout.addWidget(button)
        self.__kind.addButton(button)

        button = qt.QRadioButton(parent=panel)
        button.setText("Spiral")
        button.clicked.connect(self.generateSpiral)
        button.setCheckable(True)
        groupLayout.addWidget(button)
        self.__kind.addButton(button)

        button = qt.QRadioButton(parent=panel)
        button.setText("Gradient")
        button.clicked.connect(self.generateGradient)
        button.setCheckable(True)
        groupLayout.addWidget(button)
        self.__kind.addButton(button)

        button = qt.QRadioButton(parent=panel)
        button.setText("Composite gradient")
        button.clicked.connect(self.generateCompositeGradient)
        button.setCheckable(True)
        groupLayout.addWidget(button)
        self.__kind.addButton(button)

        button = qt.QPushButton(parent=panel)
        button.setText("Generate a new image")
        button.clicked.connect(self.generate)
        groupLayout.addWidget(button)

        # Contours

        group = qt.QGroupBox(self)
        group.setTitle("Contours")
        layout.addWidget(group)
        groupLayout = qt.QVBoxLayout(group)

        button = qt.QCheckBox(parent=panel)
        button.setText("Use the plot's mask")
        button.setCheckable(True)
        button.setChecked(True)
        button.clicked.connect(self.updateContours)
        groupLayout.addWidget(button)
        self.__useMaskButton = button

        button = qt.QPushButton(parent=panel)
        button.setText("Update contours")
        button.clicked.connect(self.updateContours)
        groupLayout.addWidget(button)

        # Implementations

        group = qt.QGroupBox(self)
        group.setTitle("Implementation")
        layout.addWidget(group)
        groupLayout = qt.QVBoxLayout(group)

        self.__impl = qt.QButtonGroup(self)
        self.__impl.setExclusive(True)

        button = qt.QRadioButton(parent=panel)
        button.setText("silx")
        button.clicked.connect(self.updateContours)
        button.setCheckable(True)
        button.setChecked(True)
        groupLayout.addWidget(button)
        self.__implMerge = button
        self.__impl.addButton(button)

        button = qt.QRadioButton(parent=panel)
        button.setText("silx with cache")
        button.clicked.connect(self.updateContours)
        button.setCheckable(True)
        groupLayout.addWidget(button)
        self.__implMergeCache = button
        self.__impl.addButton(button)

        button = qt.QRadioButton(parent=panel)
        button.setText("skimage")
        button.clicked.connect(self.updateContours)
        button.setCheckable(True)
        groupLayout.addWidget(button)
        self.__implSkimage = button
        self.__impl.addButton(button)
        if MarchingSquaresSciKitImage is None:
            button.setEnabled(False)
            button.setToolTip("skimage is not installed or not compatible")

        # Processing

        group = qt.QGroupBox(self)
        group.setTitle("Processing")
        layout.addWidget(group)
        group.setLayout(self.__createInfoLayout(group))

        # Processing

        group = qt.QGroupBox(self)
        group.setTitle("Custom level")
        layout.addWidget(group)
        groupLayout = qt.QVBoxLayout(group)

        label = qt.QLabel(parent=panel)
        self.__value = qt.QSlider(panel)
        self.__value.setOrientation(qt.Qt.Horizontal)
        self.__value.sliderMoved.connect(self.__updateCustomContours)
        self.__value.valueChanged.connect(self.__updateCustomContours)
        groupLayout.addWidget(self.__value)

        return panel

    def __createInfoLayout(self, parent):
        layout = qt.QGridLayout()

        header = qt.QLabel(parent=parent)
        header.setText("Time: ")
        label = qt.QLabel(parent=parent)
        label.setText("")
        layout.addWidget(header, 0, 0)
        layout.addWidget(label, 0, 1)
        self.__timeLabel = label

        header = qt.QLabel(parent=parent)
        header.setText("Nb polygons: ")
        label = qt.QLabel(parent=parent)
        label.setText("")
        layout.addWidget(header, 2, 0)
        layout.addWidget(label, 2, 1)
        self.__polygonsLabel = label

        header = qt.QLabel(parent=parent)
        header.setText("Nb points: ")
        label = qt.QLabel(parent=parent)
        label.setText("")
        layout.addWidget(header, 1, 0)
        layout.addWidget(label, 1, 1)
        self.__pointsLabel = label

        return layout

    def __cleanCustomContour(self):
        for name in self.__customPolygons:
            self.__plot.removeCurve(name)
        self.__customPolygons = []
        dummy = numpy.array([[[0, 0, 0, 0]]])
        item = self.__plot.getImage(legend="iso-pixels")
        item.setData([[[0, 0, 0, 0]]])

    def __cleanPolygons(self):
        for name in self.__polygons:
            self.__plot.removeCurve(name)

    def clean(self):
        self.__cleanCustomContour()
        self.__cleanPolygons()
        self.__polygons = []
        self.__image = None
        self.__mask = None

    def updateContours(self):
        self.__redrawContours()
        self.updateCustomContours()

    def __updateCustomContours(self, value):
        self.__customValue = value
        self.updateCustomContours()

    def updateCustomContours(self):
        if self.__algo is None:
            return
        value = self.__customValue
        self.__cleanCustomContour()
        if value is None:
            return

        # iso pixels
        iso_pixels = self.__algo.find_pixels(value)
        if len(iso_pixels) != 0:
            mask = numpy.zeros(self.__image.shape, dtype=numpy.int8)
            indexes = iso_pixels[:, 0] * self.__image.shape[1] + iso_pixels[:, 1]
            mask = mask.ravel()
            mask[indexes] = 1
            mask.shape = self.__image.shape
            mask = createRgbaMaskImage(mask, color=numpy.array([255, 0, 0, 128]))
            item = self.__plot.getImage(legend="iso-pixels")
            item.setData(mask)

        # iso contours
        polygons = self.__algo.find_contours(value)
        for ipolygon, polygon in enumerate(polygons):
            if len(polygon) == 0:
                continue
            isClosed = numpy.allclose(polygon[0], polygon[-1])
            x = polygon[:, 1] + 0.5
            y = polygon[:, 0] + 0.5
            legend = "custom-polygon-%d" % ipolygon
            self.__customPolygons.append(legend)
            self.__plot.addCurve(x=x, y=y, linestyle="--", color="red", linewidth=2.0, legend=legend, resetzoom=False)

    def __updateAlgo(self, image, mask=None):
        if mask is None:
            if self.__useMaskButton.isChecked():
                mask = self.__plot.getMaskToolsDockWidget().getSelectionMask()

        self.__image = image
        self.__mask = mask

        implButton = self.__impl.checkedButton()
        if implButton == self.__implMerge:
            from silx.image.marchingsquares import MarchingSquaresMergeImpl
            self.__algo = MarchingSquaresMergeImpl(self.__image, self.__mask)
        elif implButton == self.__implMergeCache:
            from silx.image.marchingsquares import MarchingSquaresMergeImpl
            self.__algo = MarchingSquaresMergeImpl(self.__image, self.__mask, use_minmax_cache=True)
        elif implButton == self.__implSkimage and MarchingSquaresSciKitImage is not None:
            self.__algo = MarchingSquaresSciKitImage(self.__image, self.__mask)
        else:
            _logger.error("No algorithm available")
            self.__algo = None

    def setData(self, image, mask=None, value=0.0):
        self.clean()

        self.__updateAlgo(image, mask=None)

        # image
        item = self.__plot.getImage(legend="image")
        item.setData(image)
        item.setColormap(self.__colormap)

        self.__plot.resetZoom()

    def __redrawContours(self):
        self.__updateAlgo(self.__image)
        if self.__algo is None:
            return
        self.__cleanPolygons()
        self.__drawContours(self.__values, self.__lineStyleCallback)

    def __drawContours(self, values, lineStyleCallback=None):
        if self.__algo is None:
            return

        self.__values = values
        self.__lineStyleCallback = lineStyleCallback
        if self.__values is None:
            return

        nbTime = 0
        nbPolygons = 0
        nbPoints = 0

        # iso contours
        ipolygon = 0
        for ivalue, value in enumerate(values):
            startTime = time.time()
            polygons = self.__algo.find_contours(value)
            nbTime += (time.time() - startTime)
            nbPolygons += len(polygons)
            for polygon in polygons:
                if len(polygon) == 0:
                    continue
                nbPoints += len(polygon)
                isClosed = numpy.allclose(polygon[0], polygon[-1])
                x = polygon[:, 1] + 0.5
                y = polygon[:, 0] + 0.5
                legend = "polygon-%d" % ipolygon
                if lineStyleCallback is not None:
                    extraStyle = lineStyleCallback(value, ivalue, ipolygon)
                else:
                    extraStyle = {"linestyle": "-", "linewidth": 1.0, "color": "black"}
                self.__polygons.append(legend)
                self.__plot.addCurve(x=x, y=y, legend=legend, resetzoom=False, **extraStyle)
                ipolygon += 1

        self.__timeLabel.setText("%0.3fs" % nbTime)
        self.__polygonsLabel.setText("%d" % nbPolygons)
        self.__pointsLabel.setText("%d" % nbPoints)

    def __defineDefaultValues(self, value=None):
        # Do not use min and max to avoid to create iso contours on small
        # and many artefacts
        if value is None:
            value = self.__image.mean()
        self.__customValue = value
        div = 12
        delta = (self.__image.max() - self.__image.min()) / div
        self.__value.setValue(int(numpy.round(value)))
        minv = self.__image.min() + delta
        maxv = self.__image.min() + delta * (div - 1)
        self.__value.setRange(int(numpy.floor(minv)), int(numpy.ceil(maxv)))
        self.updateCustomContours()

    def generate(self):
        self.__kind.checkedButton().click()

    def generateSpiral(self):
        shape = 512
        nb_spiral = numpy.random.randint(1, 8)
        freq = numpy.random.randint(2, 50)
        image = create_spiral(shape, nb_spiral, freq)
        image *= 1000.0
        self.__colormap = Colormap("cool")
        self.setData(image=image, mask=None)
        self.__defineDefaultValues()

    def generateIsland(self):
        shape = (512, 512)
        image = create_island(shape, summit=4808.72, under_water=1500)
        self.__colormap = Colormap("terrain")
        self.setData(image=image, mask=None)

        values = range(-800, 5000, 200)

        def styleCallback(value, ivalue, ipolygon):
            if value == 0:
                style = {"linestyle": "-", "linewidth": 1.0, "color": "black"}
            elif value % 1000 == 0:
                style = {"linestyle": "--", "linewidth": 0.5, "color": "black"}
            else:
                style = {"linestyle": "--", "linewidth": 0.1, "color": "black"}
            return style

        self.__drawContours(values, styleCallback)

        self.__value.setValue(0)
        self.__value.setRange(0, 5000)
        self.__updateCustomContours(0)

    def generateMagneticField(self):
        shape = 512
        x1 = numpy.random.random() * 2 - 1
        y1 = numpy.random.random() * 2 - 1
        x2 = numpy.random.random() * 2 - 1
        y2 = numpy.random.random() * 2 - 1
        image = create_magnetic_field(shape, x1, y1, x2, y2)
        self.__colormap = Colormap("coolwarm")
        self.setData(image=image, mask=None)

        maximum = abs(image.max())
        m = abs(image.min())
        if m > maximum:
            maximum = m
        maximum = int(maximum)
        values = range(-maximum, maximum, maximum // 20)

        def styleCallback(value, ivalue, ipolygon):
            if (ivalue % 2) == 0:
                style = {"linestyle": "-", "linewidth": 0.5, "color": "black"}
            else:
                style = {"linestyle": "-", "linewidth": 0.5, "color": "white"}
            return style

        self.__drawContours(values, styleCallback)
        self.__defineDefaultValues(value=0)

    def generateGravityField(self):
        shape = 512
        nb = numpy.random.randint(2, 10)
        objects = []
        for _ in range(nb):
            x = numpy.random.random() * 2 - 1
            y = numpy.random.random() * 2 - 1
            m = numpy.random.random() * 10 + 1.0
            objects.append((x, y, m))
        image = create_gravity_field(shape, objects)
        self.__colormap = Colormap("inferno")
        self.setData(image=image, mask=None)

        delta = (image.max() - image.min()) / 30.0
        values = numpy.arange(image.min(), image.max(), delta)

        def styleCallback(value, ivalue, ipolygon):
            return {"linestyle": "-", "linewidth": 0.1, "color": "white"}

        self.__drawContours(values, styleCallback)
        self.__defineDefaultValues()

    def generateGradient(self):
        shape = 512
        dx = numpy.random.random() * 2 - 1
        dy = numpy.random.random() * 2 - 1
        sx = numpy.random.randint(10, 5000) / 10.0
        sy = numpy.random.randint(10, 5000) / 10.0
        image = create_gradient(shape, dx=dx, dy=dy, sx=sx, sy=sy)
        image *= 1000.0

        def styleCallback(value, ivalue, ipolygon):
            colors = ["#9400D3", "#4B0082", "#0000FF", "#00FF00", "#FFFF00", "#FF7F00", "#FF0000"]
            color = colors[ivalue % len(colors)]
            style = {"linestyle": "-", "linewidth": 2.0, "color": color}
            return style
        delta = (image.max() - image.min()) / 9.0
        values = numpy.arange(image.min(), image.max(), delta)
        values = values[1:8]

        self.__colormap = Colormap("Greys")
        self.setData(image=image, mask=None)
        self.__drawContours(values, styleCallback)
        self.__defineDefaultValues()

    def generateCompositeGradient(self):
        shape = 512
        hole = 1 / 4.0
        dx = numpy.random.random() * hole - hole / 2.0
        dy = numpy.random.random() * hole - hole * 2
        sx = numpy.random.random() * 10.0 + 1
        sy = numpy.random.random() * 10.0 + 1
        image = create_composite_gradient(shape, dx, dy, sx, sy)
        image *= 1000.0

        def styleCallback(value, ivalue, ipolygon):
            colors = ["#9400D3", "#4B0082", "#0000FF", "#00FF00", "#FFFF00", "#FF7F00", "#FF0000"]
            color = colors[ivalue % len(colors)]
            style = {"linestyle": "-", "linewidth": 2.0, "color": color}
            return style
        delta = (image.max() - image.min()) / 9.0
        values = numpy.arange(image.min(), image.max(), delta)
        values = values[1:8]

        self.__colormap = Colormap("Greys")
        self.setData(image=image, mask=None)
        self.__drawContours(values, styleCallback)
        self.__defineDefaultValues()


def main():
    app = qt.QApplication([])
    sys.excepthook = qt.exceptionHandler
    window = FindContours()
    window.generateIsland()
    window.show()
    result = app.exec_()
    # remove ending warnings relative to QTimer
    app.deleteLater()
    return result


if __name__ == "__main__":
    result = main()
    sys.exit(result)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2014-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module implements labels layout on graph axes."""

from __future__ import absolute_import, division, unicode_literals

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "18/10/2016"


import math


# utils #######################################################################

def numberOfDigits(tickSpacing):
    """Returns the number of digits to display for text label.

    :param float tickSpacing: Step between ticks in data space.
    :return: Number of digits to show for labels.
    :rtype: int
    """
    nfrac = int(-math.floor(math.log10(tickSpacing)))
    if nfrac < 0:
        nfrac = 0
    return nfrac


# Nice Numbers ################################################################

# This is the original niceNum implementation. For the date time ticks a more
# generic implementation was needed.
#
# def _niceNum(value, isRound=False):
#     expvalue = math.floor(math.log10(value))
#     frac = value/pow(10., expvalue)
#     if isRound:
#         if frac < 1.5:
#             nicefrac = 1.
#         elif frac < 3.:    # In niceNumGeneric this is (2+5)/2 = 3.5
#             nicefrac = 2.
#         elif frac < 7.:
#             nicefrac = 5.  # In niceNumGeneric this is (5+10)/2 = 7.5
#         else:
#             nicefrac = 10.
#     else:
#         if frac <= 1.:
#             nicefrac = 1.
#         elif frac <= 2.:
#             nicefrac = 2.
#         elif frac <= 5.:
#             nicefrac = 5.
#         else:
#             nicefrac = 10.
#     return nicefrac * pow(10., expvalue)


def niceNumGeneric(value, niceFractions=None, isRound=False):
    """ A more generic implementation of the _niceNum function

    Allows the user to specify the fractions instead of using a hardcoded
    list of [1, 2, 5, 10.0].
    """
    if value == 0:
        return value

    if niceFractions is None:  # Use default values
        niceFractions = 1., 2., 5., 10.
        roundFractions = (1.5, 3., 7., 10.) if isRound else niceFractions

    else:
        roundFractions = list(niceFractions)
        if isRound:
            # Take the average with the next element. The last remains the same.
            for i in range(len(roundFractions) - 1):
                roundFractions[i] = (niceFractions[i] + niceFractions[i+1]) / 2

    highest = niceFractions[-1]
    value = float(value)

    expvalue = math.floor(math.log(value, highest))
    frac = value / pow(highest, expvalue)

    for niceFrac, roundFrac in zip(niceFractions, roundFractions):
        if frac <= roundFrac:
            return niceFrac * pow(highest, expvalue)

    # should not come here
    assert False, "should not come here"


def niceNumbers(vMin, vMax, nTicks=5):
    """Returns tick positions.

    This function implements graph labels layout using nice numbers
    by Paul Heckbert from "Graphics Gems", Academic Press, 1990.
    See `C code <http://tog.acm.org/resources/GraphicsGems/gems/Label.c>`_.

    :param float vMin: The min value on the axis
    :param float vMax: The max value on the axis
    :param int nTicks: The number of ticks to position
    :returns: min, max, increment value of tick positions and
              number of fractional digit to show
    :rtype: tuple
    """
    vrange = niceNumGeneric(vMax - vMin, isRound=False)
    spacing = niceNumGeneric(vrange / nTicks, isRound=True)
    graphmin = math.floor(vMin / spacing) * spacing
    graphmax = math.ceil(vMax / spacing) * spacing
    nfrac = numberOfDigits(spacing)
    return graphmin, graphmax, spacing, nfrac


def _frange(start, stop, step):
    """range for float (including stop)."""
    assert step >= 0.
    while start <= stop:
        yield start
        start += step


def ticks(vMin, vMax, nbTicks=5):
    """Returns tick positions and labels using nice numbers algorithm.

    This enforces ticks to be within [vMin, vMax] range.
    It returns at least 1 tick (when vMin == vMax).

    :param float vMin: The min value on the axis
    :param float vMax: The max value on the axis
    :param int nbTicks: The number of ticks to position
    :returns: tick positions and corresponding text labels
    :rtype: 2-tuple: list of float, list of string
    """
    assert vMin <= vMax
    if vMin == vMax:
        positions = [vMin]
        nfrac = 0

    else:
        start, end, step, nfrac = niceNumbers(vMin, vMax, nbTicks)
        positions = [t for t in _frange(start, end, step) if vMin <= t <= vMax]

        # Makes sure there is at least 2 ticks
        if len(positions) < 2:
            positions = [vMin, vMax]
            nfrac = numberOfDigits(vMax - vMin)

    # Generate labels
    format_ = '%g' if nfrac == 0 else '%.{}f'.format(nfrac)
    labels = [format_ % tick for tick in positions]
    return positions, labels


def niceNumbersAdaptative(vMin, vMax, axisLength, tickDensity):
    """Returns tick positions using :func:`niceNumbers` and a
    density of ticks.

    axisLength and tickDensity are based on the same unit (e.g., pixel).

    :param float vMin: The min value on the axis
    :param float vMax: The max value on the axis
    :param float axisLength: The length of the axis.
    :param float tickDensity: The density of ticks along the axis.
    :returns: min, max, increment value of tick positions and
              number of fractional digit to show
    :rtype: tuple
    """
    # At least 2 ticks
    nticks = max(2, int(round(tickDensity * axisLength)))
    tickmin, tickmax, step, nfrac = niceNumbers(vMin, vMax, nticks)

    return tickmin, tickmax, step, nfrac


# Nice Numbers for log scale ##################################################

def niceNumbersForLog10(minLog, maxLog, nTicks=5):
    """Return tick positions for logarithmic scale

    :param float minLog: log10 of the min value on the axis
    :param float maxLog: log10 of the max value on the axis
    :param int nTicks: The number of ticks to position
    :returns: log10 of min, max, increment value of tick positions and
              number of fractional digit to show
    :rtype: tuple of int
    """
    graphminlog = math.floor(minLog)
    graphmaxlog = math.ceil(maxLog)
    rangelog = graphmaxlog - graphminlog

    if rangelog <= nTicks:
        spacing = 1.
    else:
        spacing = math.floor(rangelog / nTicks)

        graphminlog = math.floor(graphminlog / spacing) * spacing
        graphmaxlog = math.ceil(graphmaxlog / spacing) * spacing

    nfrac = numberOfDigits(spacing)

    return int(graphminlog), int(graphmaxlog), int(spacing), nfrac


def niceNumbersAdaptativeForLog10(vMin, vMax, axisLength, tickDensity):
    """Returns tick positions using :func:`niceNumbers` and a
    density of ticks.

    axisLength and tickDensity are based on the same unit (e.g., pixel).

    :param float vMin: The min value on the axis
    :param float vMax: The max value on the axis
    :param float axisLength: The length of the axis.
    :param float tickDensity: The density of ticks along the axis.
    :returns: log10 of min, max, increment value of tick positions and
              number of fractional digit to show
    :rtype: tuple
    """
    # At least 2 ticks
    nticks = max(2, int(round(tickDensity * axisLength)))
    tickmin, tickmax, step, nfrac = niceNumbersForLog10(vMin, vMax, nticks)

    return tickmin, tickmax, step, nfrac


def computeLogSubTicks(ticks, lowBound, highBound):
    """Return the sub ticks for the log scale for all given ticks if subtick
    is in [lowBound, highBound]

    :param ticks: log10 of the ticks
    :param lowBound: the lower boundary of ticks
    :param highBound: the higher boundary of ticks
    :return: all the sub ticks contained in ticks (log10)
    """
    if len(ticks) < 1:
        return []

    res = []
    for logPos in ticks:
        dataOrigPos = logPos
        for index in range(2, 10):
            dataPos = dataOrigPos * index
            if lowBound <= dataPos <= highBound:
                res.append(dataPos)
    return res



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""Tests for commonh5 wrapper"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "21/09/2017"

import logging
import numpy
import unittest
import tempfile
import shutil

_logger = logging.getLogger(__name__)

import silx.io
import silx.io.utils
import h5py

try:
    from .. import commonh5
except ImportError:
    commonh5 = None


class TestCommonFeatures(unittest.TestCase):
    """Test common features supported by h5py and our implementation."""

    @classmethod
    def createFile(cls):
        return None

    @classmethod
    def setUpClass(cls):
        # Set to None cause create_resource can raise an excpetion
        cls.h5 = None
        cls.h5 = cls.create_resource()
        if cls.h5 is None:
            raise unittest.SkipTest("File not created")

    @classmethod
    def create_resource(cls):
        """Must be implemented"""
        return None

    @classmethod
    def tearDownClass(cls):
        cls.h5 = None

    def test_file(self):
        node = self.h5
        self.assertTrue(silx.io.is_file(node))
        self.assertTrue(silx.io.is_group(node))
        self.assertFalse(silx.io.is_dataset(node))
        self.assertEqual(len(node.attrs), 0)

    def test_group(self):
        node = self.h5["group"]
        self.assertFalse(silx.io.is_file(node))
        self.assertTrue(silx.io.is_group(node))
        self.assertFalse(silx.io.is_dataset(node))
        self.assertEqual(len(node.attrs), 0)
        class_ = self.h5.get("group", getclass=True)
        classlink = self.h5.get("group", getlink=True, getclass=True)
        self.assertEqual(class_, h5py.Group)
        self.assertEqual(classlink, h5py.HardLink)

    def test_dataset(self):
        node = self.h5["group/dataset"]
        self.assertFalse(silx.io.is_file(node))
        self.assertFalse(silx.io.is_group(node))
        self.assertTrue(silx.io.is_dataset(node))
        self.assertEqual(len(node.attrs), 0)
        class_ = self.h5.get("group/dataset", getclass=True)
        classlink = self.h5.get("group/dataset", getlink=True, getclass=True)
        self.assertEqual(class_, h5py.Dataset)
        self.assertEqual(classlink, h5py.HardLink)

    def test_soft_link(self):
        node = self.h5["link/soft_link"]
        self.assertEqual(node.name, "/link/soft_link")
        class_ = self.h5.get("link/soft_link", getclass=True)
        link = self.h5.get("link/soft_link", getlink=True)
        classlink = self.h5.get("link/soft_link", getlink=True, getclass=True)
        self.assertEqual(class_, h5py.Dataset)
        self.assertTrue(isinstance(link, (h5py.SoftLink, commonh5.SoftLink)))
        self.assertTrue(silx.io.utils.is_softlink(link))
        self.assertEqual(classlink, h5py.SoftLink)
 
    def test_external_link(self):
        node = self.h5["link/external_link"]
        self.assertEqual(node.name, "/target/dataset")
        class_ = self.h5.get("link/external_link", getclass=True)
        classlink = self.h5.get("link/external_link", getlink=True, getclass=True)
        self.assertEqual(class_, h5py.Dataset)
        self.assertEqual(classlink, h5py.ExternalLink)

    def test_external_link_to_link(self):
        node = self.h5["link/external_link_to_link"]
        self.assertEqual(node.name, "/target/link")
        class_ = self.h5.get("link/external_link_to_link", getclass=True)
        classlink = self.h5.get("link/external_link_to_link", getlink=True, getclass=True)
        self.assertEqual(class_, h5py.Dataset)
        self.assertEqual(classlink, h5py.ExternalLink)

    def test_create_groups(self):
        c = self.h5.create_group(self.id() + "/a/b/c")
        d = c.create_group("/" + self.id() + "/a/b/d")

        self.assertRaises(ValueError, self.h5.create_group, self.id() + "/a/b/d")
        self.assertEqual(c.name, "/" + self.id() + "/a/b/c")
        self.assertEqual(d.name, "/" + self.id() + "/a/b/d")

    def test_setitem_python_object_dataset(self):
        group = self.h5.create_group(self.id())
        group["a"] = 10
        self.assertEqual(group["a"].dtype.kind, "i")

    def test_setitem_numpy_dataset(self):
        group = self.h5.create_group(self.id())
        group["a"] = numpy.array([10, 20, 30])
        self.assertEqual(group["a"].dtype.kind, "i")
        self.assertEqual(group["a"].shape, (3,))

    def test_setitem_link(self):
        group = self.h5.create_group(self.id())
        group["a"] = 10
        group["b"] = group["a"]
        self.assertEqual(group["b"].dtype.kind, "i")

    def test_setitem_dataset_is_sub_group(self):
        self.h5[self.id() + "/a"] = 10


class TestCommonFeatures_h5py(TestCommonFeatures):
    """Check if h5py is compliant with what we expect."""

    @classmethod
    def create_resource(cls):
        cls.tmp_dir = tempfile.mkdtemp()

        externalh5 = h5py.File(cls.tmp_dir + "/external.h5", mode="w")
        externalh5["target/dataset"] = 50
        externalh5["target/link"] = h5py.SoftLink("/target/dataset")
        externalh5.close()

        h5 = h5py.File(cls.tmp_dir + "/base.h5", mode="w")
        h5["group/dataset"] = 50
        h5["link/soft_link"] = h5py.SoftLink("/group/dataset")
        h5["link/external_link"] = h5py.ExternalLink("external.h5", "/target/dataset")
        h5["link/external_link_to_link"] = h5py.ExternalLink("external.h5", "/target/link")

        return h5

    @classmethod
    def tearDownClass(cls):
        super(TestCommonFeatures_h5py, cls).tearDownClass()
        if hasattr(cls, "tmp_dir") and cls.tmp_dir is not None:
            shutil.rmtree(cls.tmp_dir)


class TestCommonFeatures_commonH5(TestCommonFeatures):
    """Check if commonh5 is compliant with h5py."""

    @classmethod
    def create_resource(cls):
        h5 = commonh5.File("base.h5", "w")
        h5.create_group("group").create_dataset("dataset", data=numpy.int32(50))

        link = h5.create_group("link")
        link.add_node(commonh5.SoftLink("soft_link", "/group/dataset"))

        return h5

    def test_external_link(self):
        # not applicable
        pass

    def test_external_link_to_link(self):
        # not applicable
        pass


class TestSpecificCommonH5(unittest.TestCase):
    """Test specific features from commonh5.

    Test of shared features should be done by TestCommonFeatures."""

    def setUp(self):
        if commonh5 is None:
            self.skipTest("silx.io.commonh5 is needed")

    def test_node_attrs(self):
        node = commonh5.Node("Foo", attrs={"a": 1})
        self.assertEqual(node.attrs["a"], 1)
        node.attrs["b"] = 8
        self.assertEqual(node.attrs["b"], 8)
        node.attrs["b"] = 2
        self.assertEqual(node.attrs["b"], 2)

    def test_node_readonly_attrs(self):
        f = commonh5.File(name="Foo", mode="r")
        node = commonh5.Node("Foo", attrs={"a": 1})
        node.attrs["b"] = 8
        f.add_node(node)
        self.assertEqual(node.attrs["b"], 8)
        try:
            node.attrs["b"] = 1
            self.fail()
        except RuntimeError:
            pass

    def test_create_dataset(self):
        f = commonh5.File(name="Foo", mode="w")
        node = f.create_dataset("foo", data=numpy.array([1]))
        self.assertIs(node.parent, f)
        self.assertIs(f["foo"], node)

    def test_create_group(self):
        f = commonh5.File(name="Foo", mode="w")
        node = f.create_group("foo")
        self.assertIs(node.parent, f)
        self.assertIs(f["foo"], node)

    def test_readonly_create_dataset(self):
        f = commonh5.File(name="Foo", mode="r")
        try:
            f.create_dataset("foo", data=numpy.array([1]))
            self.fail()
        except RuntimeError:
            pass

    def test_readonly_create_group(self):
        f = commonh5.File(name="Foo", mode="r")
        try:
            f.create_group("foo")
            self.fail()
        except RuntimeError:
            pass

    def test_create_unicode_dataset(self):
        f = commonh5.File(name="Foo", mode="w")
        try:
            f.create_dataset("foo", data=numpy.array(u"aaaa"))
            self.fail()
        except TypeError:
            pass

    def test_setitem_dataset(self):
        self.h5 = commonh5.File(name="Foo", mode="w")
        group = self.h5.create_group(self.id())
        group["a"] = commonh5.Dataset(None, data=numpy.array(10))
        self.assertEqual(group["a"].dtype.kind, "i")

    def test_setitem_explicit_link(self):
        self.h5 = commonh5.File(name="Foo", mode="w")
        group = self.h5.create_group(self.id())
        group["a"] = 10
        group["b"] = commonh5.SoftLink(None, path="/" + self.id() + "/a")
        self.assertEqual(group["b"].dtype.kind, "i")


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    test_suite = unittest.TestSuite()
    test_suite.addTest(loadTests(TestCommonFeatures_h5py))
    test_suite.addTest(loadTests(TestCommonFeatures_commonH5))
    test_suite.addTest(loadTests(TestSpecificCommonH5))
    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Functions and classes for array-like objects, implementing common numpy
array features for datasets or nested sequences, while trying to avoid copying
data.

Classes:

    - :class:`DatasetView`: Similar to a numpy view, to access
      a h5py dataset as if it was transposed, without casting it into a
      numpy array (this lets h5py handle reading the data from the
      file into memory, as needed).
    - :class:`ListOfImages`: Similar to a numpy view, to access
      a list of 2D numpy arrays as if it was a 3D array (possibly transposed),
      without casting it into a numpy array.

Functions:

    - :func:`is_array`
    - :func:`is_list_of_arrays`
    - :func:`is_nested_sequence`
    - :func:`get_shape`
    - :func:`get_dtype`
    - :func:`get_concatenated_dtype`

"""

from __future__ import absolute_import, print_function, division

import sys

import numpy
import six
import numbers

__authors__ = ["P. Knobel"]
__license__ = "MIT"
__date__ = "26/04/2017"


def is_array(obj):
    """Return True if object implements necessary attributes to be
    considered similar to a numpy array.

    Attributes needed are "shape", "dtype", "__getitem__"
    and "__array__".

    :param obj: Array-like object (numpy array, h5py dataset...)
    :return: boolean
    """
    # add more required attribute if necessary
    for attr in ("shape", "dtype", "__array__", "__getitem__"):
        if not hasattr(obj, attr):
            return False
    return True


def is_list_of_arrays(obj):
    """Return True if object is a sequence of numpy arrays,
    e.g. a list of images as 2D arrays.

    :param obj: list of arrays
    :return: boolean"""
    # object must not be a numpy array
    if is_array(obj):
        return False

    # object must have a __len__ method
    if not hasattr(obj, "__len__"):
        return False

    # all elements in sequence must be arrays
    for arr in obj:
        if not is_array(arr):
            return False

    return True


def is_nested_sequence(obj):
    """Return True if object is a nested sequence.

    A simple 1D sequence is considered to be a nested sequence.

    Numpy arrays and h5py datasets are not considered to be nested sequences.

    To test if an object is a nested sequence in a more general sense,
    including arrays and datasets, use::

        is_nested_sequence(obj) or is_array(obj)

    :param obj: nested sequence (numpy array, h5py dataset...)
    :return: boolean"""
    # object must not be a numpy array
    if is_array(obj):
        return False

    if not hasattr(obj, "__len__"):
        return False

    # obj must not be a list of (lists of) numpy arrays
    subsequence = obj
    while hasattr(subsequence, "__len__"):
        if is_array(subsequence):
            return False
        # strings cause infinite loops
        if isinstance(subsequence, six.string_types + (six.binary_type, )):
            return True
        subsequence = subsequence[0]

    # object has __len__ and is not an array
    return True


def get_shape(array_like):
    """Return shape of an array like object.

    In case the object is a nested sequence but not an array or dataset
    (list of lists, tuples...), the size of each dimension is assumed to be
    uniform, and is deduced from the length of the first sequence.

    :param array_like: Array like object: numpy array, hdf5 dataset,
        multi-dimensional sequence
    :return: Shape of array, as a tuple of integers
    """
    if hasattr(array_like, "shape"):
        return array_like.shape

    shape = []
    subsequence = array_like
    while hasattr(subsequence, "__len__"):
        shape.append(len(subsequence))
        # strings cause infinite loops
        if isinstance(subsequence, six.string_types + (six.binary_type, )):
            break
        subsequence = subsequence[0]

    return tuple(shape)


def get_dtype(array_like):
    """Return dtype of an array like object.

    In the case of a nested sequence, the type of the first value
    is inspected.

    :param array_like: Array like object: numpy array, hdf5 dataset,
        multi-dimensional nested sequence
    :return: numpy dtype of object
    """
    if hasattr(array_like, "dtype"):
        return array_like.dtype

    subsequence = array_like
    while hasattr(subsequence, "__len__"):
        # strings cause infinite loops
        if isinstance(subsequence, six.string_types + (six.binary_type, )):
            break
        subsequence = subsequence[0]

    return numpy.dtype(type(subsequence))


def get_concatenated_dtype(arrays):
    """Return dtype of array resulting of concatenation
    of a list of arrays (without actually concatenating
    them).

    :param arrays: list of numpy arrays
    :return: resulting dtype after concatenating arrays
    """
    dtypes = {a.dtype for a in arrays}
    dummy = []
    for dt in dtypes:
        dummy.append(numpy.zeros((1, 1), dtype=dt))
    return numpy.array(dummy).dtype


class ListOfImages(object):
    """This class provides a way to access values and slices in a stack of
    images stored as a list of 2D numpy arrays, without creating a 3D numpy
    array first.

    A transposition can be specified, as a 3-tuple of dimensions in the wanted
    order. For example, to transpose from ``xyz`` ``(0, 1, 2)`` into ``yzx``,
    the transposition tuple is ``(1, 2, 0)``

    All the 2D arrays in the list must have the same shape.

    The global dtype of the stack of images is the one that would be obtained
    by casting the list of 2D arrays into a 3D numpy array.

    :param images: list of 2D numpy arrays, or :class:`ListOfImages` object
    :param transposition: Tuple of dimension numbers in the wanted order
    """
    def __init__(self, images, transposition=None):
        """

        """
        super(ListOfImages, self).__init__()

        # if images is a ListOfImages instance, get the underlying data
        # as a list of 2D arrays
        if isinstance(images, ListOfImages):
            images = images.images

        # test stack of images is as expected
        assert is_list_of_arrays(images), \
            "Image stack must be a list of arrays"
        image0_shape = images[0].shape
        for image in images:
            assert image.ndim == 2, \
                "Images must be 2D numpy arrays"
            assert image.shape == image0_shape, \
                "All images must have the same shape"

        self.images = images
        """List of images"""

        self.shape = (len(images), ) + image0_shape
        """Tuple of array dimensions"""
        self.dtype = get_concatenated_dtype(images)
        """Data-type of the global array"""
        self.ndim = 3
        """Number of array dimensions"""

        self.size = len(images) * image0_shape[0] * image0_shape[1]
        """Number of elements in the array."""

        self.transposition = list(range(self.ndim))
        """List of dimension indices, in an order depending on the
        specified transposition. By default this is simply
        [0, ..., self.ndim], but it can be changed by specifying a different
        ``transposition`` parameter at initialization.

        Use :meth:`transpose`, to create a new :class:`ListOfImages`
        with a different :attr:`transposition`.
        """

        if transposition is not None:
            assert len(transposition) == self.ndim
            assert set(transposition) == set(list(range(self.ndim))), \
                "Transposition must be a sequence containing all dimensions"
            self.transposition = transposition
            self.__sort_shape()

    def __sort_shape(self):
        """Sort shape in the order defined in :attr:`transposition`
        """
        new_shape = tuple(self.shape[dim] for dim in self.transposition)
        self.shape = new_shape

    def __sort_indices(self, indices):
        """Return array indices sorted in the order needed
        to access data in the original non-transposed images.

        :param indices: Tuple of ndim indices, in the order needed
            to access the transposed view
        :return: Sorted tuple of indices, to access original data
        """
        assert len(indices) == self.ndim
        sorted_indices = tuple(idx for (_, idx) in
                               sorted(zip(self.transposition, indices)))
        return sorted_indices

    def __array__(self, dtype=None):
        """Cast the images into a numpy array, and return it.

        If a transposition has been done on this images, return
        a transposed view of a numpy array."""
        return numpy.transpose(numpy.array(self.images, dtype=dtype),
                               self.transposition)

    def __len__(self):
        return self.shape[0]

    def transpose(self, transposition=None):
        """Return a re-ordered (dimensions permutated)
        :class:`ListOfImages`.

        The returned object refers to
        the same images but with a different :attr:`transposition`.

        :param List[int] transposition: List/tuple of dimension numbers in the
            wanted order.
            If ``None`` (default), reverse the dimensions.
        :return: new :class:`ListOfImages` object
        """
        # by default, reverse the dimensions
        if transposition is None:
            transposition = list(reversed(self.transposition))

        # If this ListOfImages is already transposed, sort new transposition
        # relative to old transposition
        elif list(self.transposition) != list(range(self.ndim)):
            transposition = [self.transposition[i] for i in transposition]

        return ListOfImages(self.images,
                            transposition)

    @property
    def T(self):
        """
        Same as self.transpose()

        :return: DatasetView with dimensions reversed."""
        return self.transpose()

    def __getitem__(self, item):
        """Handle a subset of numpy indexing with regards to the dimension
        order as specified in :attr:`transposition`

        Following features are **not supported**:

            - fancy indexing using numpy arrays
            - using ellipsis objects

        :param item: Index
        :return: value or slice as a numpy array
        """
        # 1-D slicing -> n-D slicing (n=1)
        if not hasattr(item, "__len__"):
            # first dimension index is given
            item = [item]
            # following dimensions are indexed with : (all elements)
            item += [slice(None) for _i in range(self.ndim - 1)]

        # n-dimensional slicing
        if len(item) != self.ndim:
            raise IndexError(
                "N-dim slicing requires a tuple of N indices/slices. " +
                "Needed dimensions: %d" % self.ndim)

        # get list of indices sorted in the original images order
        sorted_indices = self.__sort_indices(item)
        list_idx, array_idx = sorted_indices[0], sorted_indices[1:]

        images_selection = self.images[list_idx]

        # now we must transpose the output data
        output_dimensions = []
        frozen_dimensions = []
        for i, idx in enumerate(item):
            # slices and sequences
            if not isinstance(idx, numbers.Integral):
                output_dimensions.append(self.transposition[i])
            # regular integer index
            else:
                # whenever a dimension is fixed (indexed by an integer)
                # the number of output dimension is reduced
                frozen_dimensions.append(self.transposition[i])

        # decrement output dimensions that are above frozen dimensions
        for frozen_dim in reversed(sorted(frozen_dimensions)):
            for i, out_dim in enumerate(output_dimensions):
                if out_dim > frozen_dim:
                    output_dimensions[i] -= 1

        assert (len(output_dimensions) + len(frozen_dimensions)) == self.ndim
        assert set(output_dimensions) == set(range(len(output_dimensions)))

        # single list elements selected
        if isinstance(images_selection, numpy.ndarray):
            return numpy.transpose(images_selection[array_idx],
                                   axes=output_dimensions)
        # muliple list elements selected
        else:
            # apply selection first
            output_stack = []
            for img in images_selection:
                output_stack.append(img[array_idx])
            # then cast into a numpy array, and transpose
            return numpy.transpose(numpy.array(output_stack),
                                   axes=output_dimensions)

    def min(self):
        """
        :return: Global minimum value
        """
        min_value = self.images[0].min()
        if len(self.images) > 1:
            for img in self.images[1:]:
                min_value = min(min_value, img.min())
        return min_value

    def max(self):
        """
        :return: Global maximum value
        """
        max_value = self.images[0].max()
        if len(self.images) > 1:
            for img in self.images[1:]:
                max_value = max(max_value, img.max())
        return max_value


class DatasetView(object):
    """This class provides a way to transpose a dataset without
    casting it into a numpy array. This way, the dataset in a file need not
    necessarily be integrally read into memory to view it in a different
    transposition.

    .. note::
        The performances depend a lot on the way the dataset was written
        to file. Depending on the chunking strategy, reading a complete 2D slice
        in an unfavorable direction may still require the entire dataset to
        be read from disk.

    :param dataset: h5py dataset
    :param transposition: List of dimensions sorted in the order of
        transposition (relative to the original h5py dataset)
    """
    def __init__(self, dataset, transposition=None):
        """

        """
        super(DatasetView, self).__init__()
        self.dataset = dataset
        """original dataset"""

        self.shape = dataset.shape
        """Tuple of array dimensions"""
        self.dtype = dataset.dtype
        """Data-type of the array’s element"""
        self.ndim = len(dataset.shape)
        """Number of array dimensions"""

        size = 0
        if self.ndim:
            size = 1
            for dimsize in self.shape:
                size *= dimsize
        self.size = size
        """Number of elements in the array."""

        self.transposition = list(range(self.ndim))
        """List of dimension indices, in an order depending on the
        specified transposition. By default this is simply
        [0, ..., self.ndim], but it can be changed by specifying a different
        `transposition` parameter at initialization.

        Use :meth:`transpose`, to create a new :class:`DatasetView`
        with a different :attr:`transposition`.
        """

        if transposition is not None:
            assert len(transposition) == self.ndim
            assert set(transposition) == set(list(range(self.ndim))), \
                "Transposition must be a list containing all dimensions"
            self.transposition = transposition
            self.__sort_shape()

    def __sort_shape(self):
        """Sort shape in the order defined in :attr:`transposition`
        """
        new_shape = tuple(self.shape[dim] for dim in self.transposition)
        self.shape = new_shape

    def __sort_indices(self, indices):
        """Return array indices sorted in the order needed
        to access data in the original non-transposed dataset.

        :param indices: Tuple of ndim indices, in the order needed
            to access the view
        :return: Sorted tuple of indices, to access original data
        """
        assert len(indices) == self.ndim
        sorted_indices = tuple(idx for (_, idx) in
                               sorted(zip(self.transposition, indices)))
        return sorted_indices

    def __getitem__(self, item):
        """Handle fancy indexing with regards to the dimension order as
        specified in :attr:`transposition`

        The supported fancy-indexing syntax is explained at
        http://docs.h5py.org/en/latest/high/dataset.html#fancy-indexing.

        Additional restrictions exist if the data has been transposed:

            - numpy boolean array indexing is not supported
            - ellipsis objects are not supported

        :param item: Index, possibly fancy index (must be supported by h5py)
        :return: Sliced numpy array or numpy scalar
        """
        # no transposition, let the original dataset handle indexing
        if self.transposition == list(range(self.ndim)):
            return self.dataset[item]

        # 1-D slicing: create a list of indices to switch to n-D slicing
        if not hasattr(item, "__len__"):
            # first dimension index (list index) is given
            item = [item]
            # following dimensions are indexed with slices representing all elements
            item += [slice(None) for _i in range(self.ndim - 1)]

        # n-dimensional slicing
        if len(item) != self.ndim:
            raise IndexError(
                "N-dim slicing requires a tuple of N indices/slices. " +
                "Needed dimensions: %d" % self.ndim)

        # get list of indices sorted in the original dataset order
        sorted_indices = self.__sort_indices(item)

        output_data_not_transposed = self.dataset[sorted_indices]

        # now we must transpose the output data
        output_dimensions = []
        frozen_dimensions = []
        for i, idx in enumerate(item):
            # slices and sequences
            if not isinstance(idx, int):
                output_dimensions.append(self.transposition[i])
            # regular integer index
            else:
                # whenever a dimension is fixed (indexed by an integer)
                # the number of output dimension is reduced
                frozen_dimensions.append(self.transposition[i])

        # decrement output dimensions that are above frozen dimensions
        for frozen_dim in reversed(sorted(frozen_dimensions)):
            for i, out_dim in enumerate(output_dimensions):
                if out_dim > frozen_dim:
                    output_dimensions[i] -= 1

        assert (len(output_dimensions) + len(frozen_dimensions)) == self.ndim
        assert set(output_dimensions) == set(range(len(output_dimensions)))

        return numpy.transpose(output_data_not_transposed,
                               axes=output_dimensions)

    def __array__(self, dtype=None):
        """Cast the dataset into a numpy array, and return it.

        If a transposition has been done on this dataset, return
        a transposed view of a numpy array."""
        return numpy.transpose(numpy.array(self.dataset, dtype=dtype),
                               self.transposition)

    def __len__(self):
        return self.shape[0]

    def transpose(self, transposition=None):
        """Return a re-ordered (dimensions permutated)
        :class:`DatasetView`.

        The returned object refers to
        the same dataset but with a different :attr:`transposition`.

        :param List[int] transposition: List of dimension numbers in the wanted order.
            If ``None`` (default), reverse the dimensions.
        :return: Transposed DatasetView
        """
        # by default, reverse the dimensions
        if transposition is None:
            transposition = list(reversed(self.transposition))

        # If this DatasetView is already transposed, sort new transposition
        # relative to old transposition
        elif list(self.transposition) != list(range(self.ndim)):
            transposition = [self.transposition[i] for i in transposition]

        return DatasetView(self.dataset,
                           transposition)

    @property
    def T(self):
        """
        Same as self.transpose()

        :return: DatasetView with dimensions reversed."""
        return self.transpose()



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
Python h5 module and octave h5 module have different ways to deal with
h5 files.
This module is used to make the link between octave and python using such files.
(python is using a dictionary and octave a struct )

This module provides tool to set HDF5 file for fasttomo input.

Here is an example of a simple read and write :

.. code-block:: python
    :emphasize-lines: 3,5    

    # writing a structure
    myStruct = {'MKEEP_MASK': 0.0, 'UNSHARP_SIGMA': 0.80000000000000004 }
    writer = Octaveh5().open("my_h5file", 'a')
    writer.write('mt_struct_name', myStruct)

    # reading a h5 file
    reader = Octaveh5().open("my_h5file")
    strucDict = reader.get('mt_struct_name')

.. note:: These functions depend on the `h5py <http://www.h5py.org/>`_ 
    library, which is a mandatory dependency for `silx`.

"""

import logging
logger = logging.getLogger(__name__)
import numpy as np
import h5py

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "05/10/2016"


class Octaveh5(object):
    """This class allows communication between octave and python using hdf5 format.
    """

    def __init__(self, octave_targetted_version=3.8):
        """Constructor

        :param octave_targetted_version: the version of Octave for which we want to write this hdf5 file.
        
        This is needed because for old Octave version we need to had a hack(adding one extra character)
        """
        self.file = None
        self.octave_targetted_version = octave_targetted_version

    def open(self, h5file, mode='r'):
        """Open the h5 file which has been write by octave

        :param h5file: The path of the file to read
        :param mode: the opening mode of the file :'r', 'w'...
        """
        try:
            self.file = h5py.File(h5file, mode)
            return self
        except IOError as e:
            if mode == 'a':
                reason = "\n %s: Can t find or create " % h5file
            else:
                reason = "\n %s: File not found" % h5file
            self.file = None

            logger.info(reason)
            raise e

    def get(self, struct_name):
        """Read octave equivalent structures in hdf5 file

        :param struct_name: the identification of the top level identity we want to get from an hdf5 structure
        :return: the dictionnary of the requested struct. None if can t find it
        """
        if self.file is None:
            info = "No file currently open"
            logger.info(info)
            return None

        data_dict = {}
        grr = (list(self.file[struct_name].items())[1])[1]
        try:
            gr_level2 = grr.items()
        except AttributeError:
            reason = "no gr_level2"
            logger.info(reason)
            return None

        for key, val in iter(dict(gr_level2).items()):
            data_dict[str(key)] = list(val.items())[1][1][()]

            if list(val.items())[0][1][()] != np.string_('sq_string'):
                data_dict[str(key)] = float(data_dict[str(key)])
            else:
                if list(val.items())[0][1][()] == np.string_('sq_string'):
                    # in the case the string has been stored as an nd-array of char
                    if type(data_dict[str(key)]) is np.ndarray:
                        data_dict[str(key)] = "".join(chr(item) for item in data_dict[str(key)])
                    else:
                        data_dict[str(key)] = data_dict[str(key)].decode('UTF-8')

                # In the case Octave have added an extra character at the end
                if self.octave_targetted_version < 3.8:
                    data_dict[str(key)] = data_dict[str(key)][:-1]

        return data_dict

    def write(self, struct_name, data_dict):
        """write data_dict under the group struct_name in the open hdf5 file

        :param struct_name: the identificatioon of the structure to write in the hdf5
        :param data_dict: The python dictionnary containing the informations to write
        """
        if self.file is None:
            info = "No file currently open"
            logger.info(info)
            return

        group_l1 = self.file.create_group(struct_name)
        group_l1.attrs['OCTAVE_GLOBAL'] = np.uint8(1)
        group_l1.attrs['OCTAVE_NEW_FORMAT'] = np.uint8(1)
        group_l1.create_dataset("type", data=np.string_('scalar struct'), dtype="|S14")
        group_l2 = group_l1.create_group('value')
        for ftparams in data_dict:
            group_l3 = group_l2.create_group(ftparams)
            group_l3.attrs['OCTAVE_NEW_FORMAT'] = np.uint8(1)
            if type(data_dict[ftparams]) == str:
                group_l3.create_dataset("type", (), data=np.string_('sq_string'), dtype="|S10")
                if self.octave_targetted_version < 3.8:
                    group_l3.create_dataset("value", data=np.string_(data_dict[ftparams] + '0'))
                else:
                    group_l3.create_dataset("value", data=np.string_(data_dict[ftparams]))
            else:
                group_l3.create_dataset("type", (), data=np.string_('scalar'), dtype="|S7")
                group_l3.create_dataset("value", data=data_dict[ftparams])

    def close(self):
        """Close the file after calling read function
        """
        if self.file:
            self.file.close()

    def __del__(self):
        """Destructor
        """
        self.close()



# -*- coding: utf-8 -*-
#
#    Project: silx
#             https://github.com/silx-kit/silx
#
#    Copyright (C) 2012-2016  European Synchrotron Radiation Facility, Grenoble, France
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__authors__ = ["J. Kieffer"]
__license__ = "MIT"
__date__ = "11/01/2019"

import os
import unittest
from . import test_addition
from . import test_medfilt
from . import test_backprojection
from . import test_projection
from . import test_linalg
from . import test_array_utils
from ..codec import test as test_codec
from . import test_image
from . import test_kahan
from . import test_stats
from . import test_convolution
from . import test_sparse


def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTests(test_addition.suite())
    test_suite.addTests(test_medfilt.suite())
    test_suite.addTests(test_backprojection.suite())
    test_suite.addTests(test_projection.suite())
    test_suite.addTests(test_linalg.suite())
    test_suite.addTests(test_array_utils.suite())
    test_suite.addTests(test_codec.suite())
    test_suite.addTests(test_image.suite())
    test_suite.addTests(test_kahan.suite())
    test_suite.addTests(test_stats.suite())
    test_suite.addTests(test_convolution.suite())
    test_suite.addTests(test_sparse.suite())
    # Allow to remove sift from the project
    test_base_dir = os.path.dirname(__file__)
    sift_dir = os.path.join(test_base_dir, "..", "sift")
    if os.path.exists(sift_dir):
        from ..sift import test as test_sift
        test_suite.addTests(test_sift.suite())

    return test_suite



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Tests for BoxLayoutDockWidget"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "06/03/2018"

import unittest

from silx.gui.widgets.BoxLayoutDockWidget import BoxLayoutDockWidget
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt


class TestBoxLayoutDockWidget(TestCaseQt):
    """Tests for BoxLayoutDockWidget"""

    def setUp(self):
        """Create and show a main window"""
        self.window = qt.QMainWindow()
        self.qWaitForWindowExposed(self.window)

    def tearDown(self):
        """Delete main window"""
        self.window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.window.close()
        del self.window
        self.qapp.processEvents()

    def test(self):
        """Test update of layout direction according to dock area"""
        # Create a widget with a QBoxLayout
        layout = qt.QBoxLayout(qt.QBoxLayout.LeftToRight)
        layout.addWidget(qt.QLabel('First'))
        layout.addWidget(qt.QLabel('Second'))
        widget = qt.QWidget()
        widget.setLayout(layout)

        # Add it to a BoxLayoutDockWidget
        dock = BoxLayoutDockWidget()
        dock.setWidget(widget)

        self.window.addDockWidget(qt.Qt.BottomDockWidgetArea, dock)
        self.qapp.processEvents()
        self.assertEqual(layout.direction(), qt.QBoxLayout.LeftToRight)

        self.window.addDockWidget(qt.Qt.LeftDockWidgetArea, dock)
        self.qapp.processEvents()
        self.assertEqual(layout.direction(), qt.QBoxLayout.TopToBottom)


def suite():
    loader = unittest.defaultTestLoader.loadTestsFromTestCase
    test_suite = unittest.TestSuite()
    test_suite.addTest(loader(TestBoxLayoutDockWidget))
    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This package provides a set of Qt widgets for displaying content relative to
HDF5 format.

.. note::

    This package depends on *h5py*.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "23/09/2016"


from .Hdf5TreeView import Hdf5TreeView  # noqa
from ._utils import H5Node
from ._utils import Hdf5ContextMenuEvent  # noqa
from .NexusSortFilterProxyModel import NexusSortFilterProxyModel  # noqa
from .Hdf5TreeModel import Hdf5TreeModel  # noqa

__all__ = ['Hdf5TreeView', 'H5Node', 'Hdf5ContextMenuEvent', 'NexusSortFilterProxyModel', 'Hdf5TreeModel']



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2004-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
Widget to handle regions of interest (:class:`ROI`) on curves displayed in a
:class:`PlotWindow`.

This widget is meant to work with :class:`PlotWindow`.
"""

__authors__ = ["V.A. Sole", "T. Vincent", "H. Payno"]
__license__ = "MIT"
__date__ = "13/03/2018"

from collections import OrderedDict
import logging
import os
import sys
import functools
import numpy
from silx.io import dictdump
from silx.utils import deprecation
from silx.utils.weakref import WeakMethodProxy
from silx.utils.proxy import docstring
from .. import icons, qt
from silx.math.combo import min_max
import weakref
from silx.gui.widgets.TableWidget import TableWidget
from . import items
from .items.roi import _RegionOfInterestBase


_logger = logging.getLogger(__name__)


class CurvesROIWidget(qt.QWidget):
    """
    Widget displaying a table of ROI information.

    Implements also the following behavior:

    * if the roiTable has no ROI when showing create the default ICR one

    :param parent: See :class:`QWidget`
    :param str name: The title of this widget
    """

    sigROIWidgetSignal = qt.Signal(object)
    """Signal of ROIs modifications.

    Modification information if given as a dict with an 'event' key
    providing the type of events.

    Type of events:

    - AddROI, DelROI, LoadROI and ResetROI with keys: 'roilist', 'roidict'
    - selectionChanged with keys: 'row', 'col' 'roi', 'key', 'colheader',
      'rowheader'
    """

    sigROISignal = qt.Signal(object)

    def __init__(self, parent=None, name=None, plot=None):
        super(CurvesROIWidget, self).__init__(parent)
        if name is not None:
            self.setWindowTitle(name)
        self.__lastSigROISignal = None
        """Store the last value emitted for the sigRoiSignal. In the case the
        active curve change we need to add this extra step in order to make
        sure we won't send twice the sigROISignal.
        This come from the fact sigROISignal is connected to the 
        activeROIChanged signal which is emitted when raw and net counts
        values are changing but are not embed in the sigROISignal.
        """
        assert plot is not None
        self._plotRef = weakref.ref(plot)
        self._showAllMarkers = False
        self.currentROI = None

        layout = qt.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        self.headerLabel = qt.QLabel(self)
        self.headerLabel.setAlignment(qt.Qt.AlignHCenter)
        self.setHeader()
        layout.addWidget(self.headerLabel)

        widgetAllCheckbox = qt.QWidget(parent=self)
        self._showAllCheckBox = qt.QCheckBox("show all ROI",
                                             parent=widgetAllCheckbox)
        widgetAllCheckbox.setLayout(qt.QHBoxLayout())
        spacer = qt.QWidget(parent=widgetAllCheckbox)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Fixed)
        widgetAllCheckbox.layout().addWidget(spacer)
        widgetAllCheckbox.layout().addWidget(self._showAllCheckBox)
        layout.addWidget(widgetAllCheckbox)

        self.roiTable = ROITable(self, plot=plot)
        rheight = self.roiTable.horizontalHeader().sizeHint().height()
        self.roiTable.setMinimumHeight(4 * rheight)
        layout.addWidget(self.roiTable)
        self._roiFileDir = qt.QDir.home().absolutePath()
        self._showAllCheckBox.toggled.connect(self.roiTable.showAllMarkers)

        hbox = qt.QWidget(self)
        hboxlayout = qt.QHBoxLayout(hbox)
        hboxlayout.setContentsMargins(0, 0, 0, 0)
        hboxlayout.setSpacing(0)

        hboxlayout.addStretch(0)

        self.addButton = qt.QPushButton(hbox)
        self.addButton.setText("Add ROI")
        self.addButton.setToolTip('Create a new ROI')
        self.delButton = qt.QPushButton(hbox)
        self.delButton.setText("Delete ROI")
        self.addButton.setToolTip('Remove the selected ROI')
        self.resetButton = qt.QPushButton(hbox)
        self.resetButton.setText("Reset")
        self.addButton.setToolTip('Clear all created ROIs. We only let the '
                                  'default ROI')

        hboxlayout.addWidget(self.addButton)
        hboxlayout.addWidget(self.delButton)
        hboxlayout.addWidget(self.resetButton)

        hboxlayout.addStretch(0)

        self.loadButton = qt.QPushButton(hbox)
        self.loadButton.setText("Load")
        self.loadButton.setToolTip('Load ROIs from a .ini file')
        self.saveButton = qt.QPushButton(hbox)
        self.saveButton.setText("Save")
        self.loadButton.setToolTip('Save ROIs to a .ini file')
        hboxlayout.addWidget(self.loadButton)
        hboxlayout.addWidget(self.saveButton)
        layout.setStretchFactor(self.headerLabel, 0)
        layout.setStretchFactor(self.roiTable, 1)
        layout.setStretchFactor(hbox, 0)

        layout.addWidget(hbox)

        # Signal / Slot connections
        self.addButton.clicked.connect(self._add)
        self.delButton.clicked.connect(self._del)
        self.resetButton.clicked.connect(self._reset)

        self.loadButton.clicked.connect(self._load)
        self.saveButton.clicked.connect(self._save)

        self.roiTable.activeROIChanged.connect(self._emitCurrentROISignal)

        self._isConnected = False  # True if connected to plot signals
        self._isInit = False

        # expose API
        self.getROIListAndDict = self.roiTable.getROIListAndDict

    def getPlotWidget(self):
        """Returns the associated PlotWidget or None

        :rtype: Union[~silx.gui.plot.PlotWidget,None]
        """
        return None if self._plotRef is None else self._plotRef()

    def showEvent(self, event):
        self._visibilityChangedHandler(visible=True)
        qt.QWidget.showEvent(self, event)

    @property
    def roiFileDir(self):
        """The directory from which to load/save ROI from/to files."""
        if not os.path.isdir(self._roiFileDir):
            self._roiFileDir = qt.QDir.home().absolutePath()
        return self._roiFileDir

    @roiFileDir.setter
    def roiFileDir(self, roiFileDir):
        self._roiFileDir = str(roiFileDir)

    def setRois(self, rois, order=None):
        return self.roiTable.setRois(rois, order)

    def getRois(self, order=None):
        return self.roiTable.getRois(order)

    def setMiddleROIMarkerFlag(self, flag=True):
        return self.roiTable.setMiddleROIMarkerFlag(flag)

    def _add(self):
        """Add button clicked handler"""
        def getNextRoiName():
            rois = self.roiTable.getRois(order=None)
            roisNames = []
            [roisNames.append(roiName) for roiName in rois]
            nrois = len(rois)
            if nrois == 0:
                return "ICR"
            else:
                i = 1
                newroi = "newroi %d" % i
                while newroi in roisNames:
                    i += 1
                    newroi = "newroi %d" % i
                return newroi
        roi = ROI(name=getNextRoiName())

        if roi.getName() == "ICR":
            roi.setType("Default")
        else:
            roi.setType(self.getPlotWidget().getXAxis().getLabel())

        xmin, xmax = self.getPlotWidget().getXAxis().getLimits()
        fromdata = xmin + 0.25 * (xmax - xmin)
        todata = xmin + 0.75 * (xmax - xmin)
        if roi.isICR():
            fromdata, dummy0, todata, dummy1 = self._getAllLimits()
        roi.setFrom(fromdata)
        roi.setTo(todata)
        self.roiTable.addRoi(roi)

        # back compatibility pymca roi signals
        ddict = {}
        ddict['event'] = "AddROI"
        ddict['roilist'] = self.roiTable.roidict.values()
        ddict['roidict'] = self.roiTable.roidict
        self.sigROIWidgetSignal.emit(ddict)
        # end back compatibility pymca roi signals

    def _del(self):
        """Delete button clicked handler"""
        self.roiTable.deleteActiveRoi()

        # back compatibility pymca roi signals
        ddict = {}
        ddict['event'] = "DelROI"
        ddict['roilist'] = self.roiTable.roidict.values()
        ddict['roidict'] = self.roiTable.roidict
        self.sigROIWidgetSignal.emit(ddict)
        # end back compatibility pymca roi signals

    def _reset(self):
        """Reset button clicked handler"""
        self.roiTable.clear()
        old = self.blockSignals(True)  # avoid several sigROISignal emission
        self._add()
        self.blockSignals(old)

        # back compatibility pymca roi signals
        ddict = {}
        ddict['event'] = "ResetROI"
        ddict['roilist'] = self.roiTable.roidict.values()
        ddict['roidict'] = self.roiTable.roidict
        self.sigROIWidgetSignal.emit(ddict)
        # end back compatibility pymca roi signals

    def _load(self):
        """Load button clicked handler"""
        dialog = qt.QFileDialog(self)
        dialog.setNameFilters(
            ['INI File  *.ini', 'JSON File *.json', 'All *.*'])
        dialog.setFileMode(qt.QFileDialog.ExistingFile)
        dialog.setDirectory(self.roiFileDir)
        if not dialog.exec_():
            dialog.close()
            return

        # pyflakes bug http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=666494
        outputFile = dialog.selectedFiles()[0]
        dialog.close()

        self.roiFileDir = os.path.dirname(outputFile)
        self.roiTable.load(outputFile)

        # back compatibility pymca roi signals
        ddict = {}
        ddict['event'] = "LoadROI"
        ddict['roilist'] = self.roiTable.roidict.values()
        ddict['roidict'] = self.roiTable.roidict
        self.sigROIWidgetSignal.emit(ddict)
        # end back compatibility pymca roi signals

    def load(self, filename):
        """Load ROI widget information from a file storing a dict of ROI.

        :param str filename: The file from which to load ROI
        """
        self.roiTable.load(filename)

    def _save(self):
        """Save button clicked handler"""
        dialog = qt.QFileDialog(self)
        dialog.setNameFilters(['INI File  *.ini', 'JSON File *.json'])
        dialog.setFileMode(qt.QFileDialog.AnyFile)
        dialog.setAcceptMode(qt.QFileDialog.AcceptSave)
        dialog.setDirectory(self.roiFileDir)
        if not dialog.exec_():
            dialog.close()
            return

        outputFile = dialog.selectedFiles()[0]
        extension = '.' + dialog.selectedNameFilter().split('.')[-1]
        dialog.close()

        if not outputFile.endswith(extension):
            outputFile += extension

        if os.path.exists(outputFile):
            try:
                os.remove(outputFile)
            except IOError:
                msg = qt.QMessageBox(self)
                msg.setIcon(qt.QMessageBox.Critical)
                msg.setText("Input Output Error: %s" % (sys.exc_info()[1]))
                msg.exec_()
                return
        self.roiFileDir = os.path.dirname(outputFile)
        self.save(outputFile)

    def save(self, filename):
        """Save current ROIs of the widget as a dict of ROI to a file.

        :param str filename: The file to which to save the ROIs
        """
        self.roiTable.save(filename)

    def setHeader(self, text='ROIs'):
        """Set the header text of this widget"""
        self.headerLabel.setText("<b>%s<\b>" % text)

    @deprecation.deprecated(replacement="calculateRois",
                            reason="CamelCase convention",
                            since_version="0.7")
    def calculateROIs(self, *args, **kw):
        self.calculateRois(*args, **kw)

    def calculateRois(self, roiList=None, roiDict=None):
        """Compute ROI information"""
        return self.roiTable.calculateRois()

    def showAllMarkers(self, _show=True):
        self.roiTable.showAllMarkers(_show)

    def _getAllLimits(self):
        """Retrieve the limits based on the curves."""
        plot = self.getPlotWidget()
        curves = () if plot is None else plot.getAllCurves()
        if not curves:
            return 1.0, 1.0, 100., 100.

        xmin, ymin = None, None
        xmax, ymax = None, None

        for curve in curves:
            x = curve.getXData(copy=False)
            y = curve.getYData(copy=False)
            if xmin is None:
                xmin = x.min()
            else:
                xmin = min(xmin, x.min())
            if xmax is None:
                xmax = x.max()
            else:
                xmax = max(xmax, x.max())
            if ymin is None:
                ymin = y.min()
            else:
                ymin = min(ymin, y.min())
            if ymax is None:
                ymax = y.max()
            else:
                ymax = max(ymax, y.max())

        return xmin, ymin, xmax, ymax

    def showEvent(self, event):
        self._visibilityChangedHandler(visible=True)
        qt.QWidget.showEvent(self, event)

    def hideEvent(self, event):
        self._visibilityChangedHandler(visible=False)
        qt.QWidget.hideEvent(self, event)

    def _visibilityChangedHandler(self, visible):
        """Handle widget's visibility updates.

        It is connected to plot signals only when visible.
        """
        if visible:
            # if no ROI existing yet, add the default one
            if self.roiTable.rowCount() == 0:
                old = self.blockSignals(True)  # avoid several sigROISignal emission
                self._add()
                self.blockSignals(old)
                self.calculateRois()

    def fillFromROIDict(self, *args, **kwargs):
        self.roiTable.fillFromROIDict(*args, **kwargs)

    def _emitCurrentROISignal(self):
        ddict = {}
        ddict['event'] = "currentROISignal"
        if self.roiTable.activeRoi is not None:
            ddict['ROI'] = self.roiTable.activeRoi.toDict()
            ddict['current'] = self.roiTable.activeRoi.getName()
        else:
            ddict['current'] = None

        if self.__lastSigROISignal != ddict:
            self.__lastSigROISignal = ddict
            self.sigROISignal.emit(ddict)

    @property
    def currentRoi(self):
        return self.roiTable.activeRoi


class _FloatItem(qt.QTableWidgetItem):
    """
    Simple QTableWidgetItem overloading the < operator to deal with ordering
    """
    def __init__(self):
        qt.QTableWidgetItem.__init__(self, type=qt.QTableWidgetItem.Type)

    def __lt__(self, other):
        if self.text() in ('', ROITable.INFO_NOT_FOUND):
            return False
        if other.text() in ('', ROITable.INFO_NOT_FOUND):
            return True
        return float(self.text()) < float(other.text())


class ROITable(TableWidget):
    """Table widget displaying ROI information.

    See :class:`QTableWidget` for constructor arguments.

    Behavior: listen at the active curve changed only when the widget is
    visible. Otherwise won't compute the row and net counts...
    """

    activeROIChanged = qt.Signal()
    """Signal emitted when the active roi changed or when the value of the
    active roi are changing"""

    COLUMNS_INDEX = OrderedDict([
        ('ID', 0),
        ('ROI', 1),
        ('Type', 2),
        ('From', 3),
        ('To', 4),
        ('Raw Counts', 5),
        ('Net Counts', 6),
        ('Raw Area', 7),
        ('Net Area', 8),
    ])

    COLUMNS = list(COLUMNS_INDEX.keys())

    INFO_NOT_FOUND = '????????'

    def __init__(self, parent=None, plot=None, rois=None):
        super(ROITable, self).__init__(parent)
        self._showAllMarkers = False
        self._userIsEditingRoi = False
        """bool used to avoid conflict when editing the ROI object"""
        self._isConnected = False
        self._roiToItems = {}
        self._roiDict = {}
        """dict of ROI object. Key is ROi id, value is the ROI object"""
        self._markersHandler = _RoiMarkerManager()

        """
        Associate for each marker legend used when the `_showAllMarkers` option
        is active a roi.
        """
        self.setColumnCount(len(self.COLUMNS))
        self.setPlot(plot)
        self.__setTooltip()
        self.setSortingEnabled(True)
        self.itemChanged.connect(self._itemChanged)

    @property
    def roidict(self):
        return self._getRoiDict()

    @property
    def activeRoi(self):
        return self._markersHandler._activeRoi

    def _getRoiDict(self):
        ddict = {}
        for id in self._roiDict:
            ddict[self._roiDict[id].getName()] = self._roiDict[id]
        return ddict

    def clear(self):
        """
        .. note:: clear the interface only. keep the roidict...
        """
        self._markersHandler.clear()
        self._roiToItems = {}
        self._roiDict = {}

        qt.QTableWidget.clear(self)
        self.setRowCount(0)
        self.setHorizontalHeaderLabels(self.COLUMNS)
        header = self.horizontalHeader()
        if hasattr(header, 'setSectionResizeMode'):  # Qt5
            header.setSectionResizeMode(qt.QHeaderView.ResizeToContents)
        else:  # Qt4
            header.setResizeMode(qt.QHeaderView.ResizeToContents)
        self.sortByColumn(0, qt.Qt.AscendingOrder)
        self.hideColumn(self.COLUMNS_INDEX['ID'])

    def setPlot(self, plot):
        self.clear()
        self.plot = plot

    def __setTooltip(self):
        self.horizontalHeaderItem(self.COLUMNS_INDEX['ROI']).setToolTip(
            'Region of interest identifier')
        self.horizontalHeaderItem(self.COLUMNS_INDEX['Type']).setToolTip(
            'Type of the ROI')
        self.horizontalHeaderItem(self.COLUMNS_INDEX['From']).setToolTip(
            'X-value of the min point')
        self.horizontalHeaderItem(self.COLUMNS_INDEX['To']).setToolTip(
            'X-value of the max point')
        self.horizontalHeaderItem(self.COLUMNS_INDEX['Raw Counts']).setToolTip(
            'Estimation of the integral between y=0 and the selected curve')
        self.horizontalHeaderItem(self.COLUMNS_INDEX['Net Counts']).setToolTip(
            'Estimation of the integral between the segment [maxPt, minPt] '
            'and the selected curve')

    def setRois(self, rois, order=None):
        """Set the ROIs by providing a dictionary of ROI information.

        The dictionary keys are the ROI names.
        Each value is a sub-dictionary of ROI info with the following fields:

        - ``"from"``: x coordinate of the left limit, as a float
        - ``"to"``: x coordinate of the right limit, as a float
        - ``"type"``: type of ROI, as a string (e.g "channels", "energy")


        :param roidict: Dictionary of ROIs
        :param str order: Field used for ordering the ROIs.
             One of "from", "to", "type".
             None (default) for no ordering, or same order as specified
             in parameter ``roidict`` if provided as an OrderedDict.
        """
        assert order in [None, "from", "to", "type"]
        self.clear()

        # backward compatibility since 0.10.0
        if isinstance(rois, dict):
            for roiName, roi in rois.items():
                if isinstance(roi, ROI):
                    _roi = roi
                else:
                    roi['name'] = roiName
                    _roi = ROI._fromDict(roi)
                self.addRoi(_roi)
        else:
            for roi in rois:
                assert isinstance(roi, ROI)
                self.addRoi(roi)
        self._updateMarkers()

    def addRoi(self, roi):
        """

        :param :class:`ROI` roi: roi to add to the table
        """
        assert isinstance(roi, ROI)
        self._getItem(name='ID', row=None, roi=roi)
        self._roiDict[roi.getID()] = roi
        self._markersHandler.add(roi, _RoiMarkerHandler(roi, self.plot))
        self._updateRoiInfo(roi.getID())
        callback = functools.partial(WeakMethodProxy(self._updateRoiInfo),
                                     roi.getID())
        roi.sigChanged.connect(callback)
        # set it as the active one
        self.setActiveRoi(roi)

    def _getItem(self, name, row, roi):
        if row:
            item = self.item(row, self.COLUMNS_INDEX[name])
        else:
            item = None
        if item:
            return item
        else:
            if name == 'ID':
                assert roi
                if roi.getID() in self._roiToItems:
                    return self._roiToItems[roi.getID()]
                else:
                    # create a new row
                    row = self.rowCount()
                    self.setRowCount(self.rowCount() + 1)
                    item = qt.QTableWidgetItem(str(roi.getID()),
                                               type=qt.QTableWidgetItem.Type)
                    self._roiToItems[roi.getID()] = item
            elif name == 'ROI':
                item = qt.QTableWidgetItem(roi.getName() if roi else '',
                                           type=qt.QTableWidgetItem.Type)
                if roi.getName().upper() in ('ICR', 'DEFAULT'):
                    item.setFlags(qt.Qt.ItemIsSelectable | qt.Qt.ItemIsEnabled)
                else:
                    item.setFlags(qt.Qt.ItemIsSelectable |
                                  qt.Qt.ItemIsEnabled |
                                  qt.Qt.ItemIsEditable)
            elif name == 'Type':
                item = qt.QTableWidgetItem(type=qt.QTableWidgetItem.Type)
                item.setFlags((qt.Qt.ItemIsSelectable | qt.Qt.ItemIsEnabled))
            elif name in ('To', 'From'):
                item = _FloatItem()
                if roi.getName().upper() in ('ICR', 'DEFAULT'):
                    item.setFlags(qt.Qt.ItemIsSelectable | qt.Qt.ItemIsEnabled)
                else:
                    item.setFlags(qt.Qt.ItemIsSelectable |
                                  qt.Qt.ItemIsEnabled |
                                  qt.Qt.ItemIsEditable)
            elif name in ('Raw Counts', 'Net Counts', 'Raw Area', 'Net Area'):
                item = _FloatItem()
                item.setFlags((qt.Qt.ItemIsSelectable | qt.Qt.ItemIsEnabled))
            else:
                raise ValueError('item type not recognized')

            self.setItem(row, self.COLUMNS_INDEX[name], item)
            return item

    def _itemChanged(self, item):
        def getRoi():
            IDItem = self.item(item.row(), self.COLUMNS_INDEX['ID'])
            assert IDItem
            id = int(IDItem.text())
            assert id in self._roiDict
            roi = self._roiDict[id]
            return roi

        def signalChanged(roi):
            if self.activeRoi and roi.getID() == self.activeRoi.getID():
                self.activeROIChanged.emit()

        self._userIsEditingRoi = True
        if item.column() in (self.COLUMNS_INDEX['To'], self.COLUMNS_INDEX['From']):
            roi = getRoi()

            if item.text() not in ('', self.INFO_NOT_FOUND):
                try:
                    value = float(item.text())
                except ValueError:
                    value = 0
                changed = False
                if item.column() == self.COLUMNS_INDEX['To']:
                    if value != roi.getTo():
                        roi.setTo(value)
                        changed = True
                else:
                    assert(item.column() == self.COLUMNS_INDEX['From'])
                    if value != roi.getFrom():
                        roi.setFrom(value)
                        changed = True
                if changed:
                    self._updateMarker(roi.getName())
                    signalChanged(roi)

        if item.column() is self.COLUMNS_INDEX['ROI']:
            roi = getRoi()
            if roi.getName() != item.text():
                roi.setName(item.text())
                self._markersHandler.getMarkerHandler(roi.getID()).updateTexts()
                signalChanged(roi)

        self._userIsEditingRoi = False

    def deleteActiveRoi(self):
        """
        remove the current active roi
        """
        activeItems = self.selectedItems()
        if len(activeItems) == 0:
            return
        old = self.blockSignals(True)  # avoid several emission of sigROISignal
        roiToRm = set()
        for item in activeItems:
            row = item.row()
            itemID = self.item(row, self.COLUMNS_INDEX['ID'])
            roiToRm.add(self._roiDict[int(itemID.text())])
        [self.removeROI(roi) for roi in roiToRm]
        self.blockSignals(old)
        self.setActiveRoi(None)

    def removeROI(self, roi):
        """
        remove the requested roi

        :param str name: the name of the roi to remove from the table
        """
        if roi and roi.getID() in self._roiToItems:
            item = self._roiToItems[roi.getID()]
            self.removeRow(item.row())
            del self._roiToItems[roi.getID()]

            assert roi.getID() in self._roiDict
            del self._roiDict[roi.getID()]
            self._markersHandler.remove(roi)

            callback = functools.partial(WeakMethodProxy(self._updateRoiInfo),
                                         roi.getID())
            roi.sigChanged.connect(callback)

    def setActiveRoi(self, roi):
        """
        Define the given roi as the active one.

        .. warning:: this roi should already be registred / added to the table

        :param :class:`ROI` roi: the roi to defined as active
        """
        if roi is None:
            self.clearSelection()
            self._markersHandler.setActiveRoi(None)
            self.activeROIChanged.emit()
        else:
            assert isinstance(roi, ROI)
            if roi and roi.getID() in self._roiToItems.keys():
                # avoid several call back to setActiveROI
                old = self.blockSignals(True)
                self.selectRow(self._roiToItems[roi.getID()].row())
                self.blockSignals(old)
                self._markersHandler.setActiveRoi(roi)
                self.activeROIChanged.emit()

    def _updateRoiInfo(self, roiID):
        if self._userIsEditingRoi is True:
            return
        if roiID not in self._roiDict:
            return
        roi = self._roiDict[roiID]
        if roi.isICR():
            activeCurve = self.plot.getActiveCurve()
            if activeCurve:
                xData = activeCurve.getXData()
                if len(xData) > 0:
                    min, max = min_max(xData)
                    roi.blockSignals(True)
                    roi.setFrom(min)
                    roi.setTo(max)
                    roi.blockSignals(False)

        itemID = self._getItem(name='ID', roi=roi, row=None)
        itemName = self._getItem(name='ROI', row=itemID.row(), roi=roi)
        itemName.setText(roi.getName())

        itemType = self._getItem(name='Type', row=itemID.row(), roi=roi)
        itemType.setText(roi.getType() or self.INFO_NOT_FOUND)

        itemFrom = self._getItem(name='From', row=itemID.row(), roi=roi)
        fromdata = str(roi.getFrom()) if roi.getFrom() is not None else self.INFO_NOT_FOUND
        itemFrom.setText(fromdata)

        itemTo = self._getItem(name='To', row=itemID.row(), roi=roi)
        todata = str(roi.getTo()) if roi.getTo() is not None else self.INFO_NOT_FOUND
        itemTo.setText(todata)

        rawCounts, netCounts = roi.computeRawAndNetCounts(
            curve=self.plot.getActiveCurve(just_legend=False))
        itemRawCounts = self._getItem(name='Raw Counts', row=itemID.row(),
                                      roi=roi)
        rawCounts = str(rawCounts) if rawCounts is not None else self.INFO_NOT_FOUND
        itemRawCounts.setText(rawCounts)

        itemNetCounts = self._getItem(name='Net Counts', row=itemID.row(),
                                      roi=roi)
        netCounts = str(netCounts) if netCounts is not None else self.INFO_NOT_FOUND
        itemNetCounts.setText(netCounts)

        rawArea, netArea = roi.computeRawAndNetArea(
            curve=self.plot.getActiveCurve(just_legend=False))
        itemRawArea = self._getItem(name='Raw Area', row=itemID.row(),
                                      roi=roi)
        rawArea = str(rawArea) if rawArea is not None else self.INFO_NOT_FOUND
        itemRawArea.setText(rawArea)

        itemNetArea = self._getItem(name='Net Area', row=itemID.row(),
                                    roi=roi)
        netArea = str(netArea) if netArea is not None else self.INFO_NOT_FOUND
        itemNetArea.setText(netArea)

        if self.activeRoi and roi.getID() == self.activeRoi.getID():
            self.activeROIChanged.emit()

    def currentChanged(self, current, previous):
        if previous and current.row() != previous.row() and current.row() >= 0:
            roiItem = self.item(current.row(),
                                self.COLUMNS_INDEX['ID'])

            assert roiItem
            self.setActiveRoi(self._roiDict[int(roiItem.text())])
            self._markersHandler.updateAllMarkers()
        qt.QTableWidget.currentChanged(self, current, previous)

    @deprecation.deprecated(reason="Removed",
                            replacement="roidict and roidict.values()",
                            since_version="0.10.0")
    def getROIListAndDict(self):
        """

        :return: the list of roi objects and the dictionary of roi name to roi
                 object.
        """
        roidict = self._roiDict
        return list(roidict.values()), roidict

    def calculateRois(self, roiList=None, roiDict=None):
        """
        Update values of all registred rois (raw and net counts in particular)

        :param roiList: deprecated parameter
        :param roiDict: deprecated parameter
        """
        if roiDict:
            deprecation.deprecated_warning(name='roiDict', type_='Parameter',
                                           reason='Unused parameter',
                                           since_version="0.10.0")
        if roiList:
            deprecation.deprecated_warning(name='roiList', type_='Parameter',
                                           reason='Unused parameter',
                                           since_version="0.10.0")

        for roiID in self._roiDict:
            self._updateRoiInfo(roiID)

    def _updateMarker(self, roiID):
        """Make sure the marker of the given roi name is updated"""
        if self._showAllMarkers or (self.activeRoi
                                    and self.activeRoi.getName() == roiID):
            self._updateMarkers()

    def _updateMarkers(self):
        if self._showAllMarkers is True:
            self._markersHandler.updateMarkers()
        else:
            if not self.activeRoi or not self.plot:
                return
            assert isinstance(self.activeRoi, ROI)
            markerHandler = self._markersHandler.getMarkerHandler(self.activeRoi.getID())
            if markerHandler is not None:
                markerHandler.updateMarkers()

    def getRois(self, order):
        """
        Return the currently defined ROIs, as an ordered dict.

        The dictionary keys are the ROI names.
        Each value is a :class:`ROI` object..

        :param order: Field used for ordering the ROIs.
             One of "from", "to", "type", "netcounts", "rawcounts".
             None (default) to get the same order as displayed in the widget.
        :return: Ordered dictionary of ROI information
        """

        if order is None or order.lower() == "none":
            ordered_roilist = list(self._roiDict.values())
            res = OrderedDict([(roi.getName(), self._roiDict[roi.getID()]) for roi in ordered_roilist])
        else:
            assert order in ["from", "to", "type", "netcounts", "rawcounts"]
            ordered_roilist = sorted(self._roiDict.keys(),
                                     key=lambda roi_id: self._roiDict[roi_id].get(order))
            res = OrderedDict([(roi.getName(), self._roiDict[id]) for id in ordered_roilist])

        return res

    def save(self, filename):
        """
        Save current ROIs of the widget as a dict of ROI to a file.

        :param str filename: The file to which to save the ROIs
        """
        roilist = []
        roidict = {}
        for roiID, roi in self._roiDict.items():
            roilist.append(roi.toDict())
            roidict[roi.getName()] = roi.toDict()
        datadict = {'ROI': {'roilist': roilist, 'roidict': roidict}}
        dictdump.dump(datadict, filename)

    def load(self, filename):
        """
        Load ROI widget information from a file storing a dict of ROI.

        :param str filename: The file from which to load ROI
        """
        roisDict = dictdump.load(filename)
        rois = []

        # Remove rawcounts and netcounts from ROIs
        for roiDict in roisDict['ROI']['roidict'].values():
            roiDict.pop('rawcounts', None)
            roiDict.pop('netcounts', None)
            rois.append(ROI._fromDict(roiDict))

        self.setRois(rois)

    def showAllMarkers(self, _show=True):
        """

        :param bool _show: if true show all the markers of all the ROIs
                           boundaries otherwise will only show the one of
                           the active ROI.
        """
        self._markersHandler.setShowAllMarkers(_show)

    def setMiddleROIMarkerFlag(self, flag=True):
        """
        Activate or deactivate middle marker.

        This allows shifting both min and max limits at once, by dragging
        a marker located in the middle.

        :param bool flag: True to activate middle ROI marker
        """
        self._markersHandler._middleROIMarkerFlag = flag

    def _handleROIMarkerEvent(self, ddict):
        """Handle plot signals related to marker events."""
        if ddict['event'] == 'markerMoved':
            label = ddict['label']
            roiID = self._markersHandler.getRoiID(markerID=label)
            if roiID is not None:
                # avoid several emission of sigROISignal
                old = self.blockSignals(True)
                self._markersHandler.changePosition(markerID=label,
                                                    x=ddict['x'])
                self.blockSignals(old)
                self._updateRoiInfo(roiID)

    def showEvent(self, event):
        self._visibilityChangedHandler(visible=True)
        qt.QWidget.showEvent(self, event)

    def hideEvent(self, event):
        self._visibilityChangedHandler(visible=False)
        qt.QWidget.hideEvent(self, event)

    def _visibilityChangedHandler(self, visible):
        """Handle widget's visibility updates.

        It is connected to plot signals only when visible.
        """
        if visible:
            assert self.plot
            if self._isConnected is False:
                self.plot.sigPlotSignal.connect(self._handleROIMarkerEvent)
                self.plot.sigActiveCurveChanged.connect(self._activeCurveChanged)
                self._isConnected = True
            self.calculateRois()
        else:
            if self._isConnected:
                self.plot.sigPlotSignal.disconnect(self._handleROIMarkerEvent)
                self.plot.sigActiveCurveChanged.disconnect(self._activeCurveChanged)
                self._isConnected = False

    def _activeCurveChanged(self, curve):
        self.calculateRois()

    def setCountsVisible(self, visible):
        """
        Display the columns relative to areas or not

        :param bool visible: True if the columns 'Raw Area' and 'Net Area'
                             should be visible.
        """
        if visible is True:
            self.showColumn(self.COLUMNS_INDEX['Raw Counts'])
            self.showColumn(self.COLUMNS_INDEX['Net Counts'])
        else:
            self.hideColumn(self.COLUMNS_INDEX['Raw Counts'])
            self.hideColumn(self.COLUMNS_INDEX['Net Counts'])

    def setAreaVisible(self, visible):
        """
        Display the columns relative to areas or not

        :param bool visible: True if the columns 'Raw Area' and 'Net Area'
                             should be visible.
        """
        if visible is True:
            self.showColumn(self.COLUMNS_INDEX['Raw Area'])
            self.showColumn(self.COLUMNS_INDEX['Net Area'])
        else:
            self.hideColumn(self.COLUMNS_INDEX['Raw Area'])
            self.hideColumn(self.COLUMNS_INDEX['Net Area'])

    def fillFromROIDict(self, roilist=(), roidict=None, currentroi=None):
        """
        This function API is kept for compatibility.
        But `setRois` should be preferred.

        Set the ROIs by providing a list of ROI names and a dictionary
        of ROI information for each ROI.
        The ROI names must match an existing dictionary key.
        The name list is used to provide an order for the ROIs.
        The dictionary's values are sub-dictionaries containing 3
        mandatory fields:

        - ``"from"``: x coordinate of the left limit, as a float
        - ``"to"``: x coordinate of the right limit, as a float
        - ``"type"``: type of ROI, as a string (e.g "channels", "energy")

        :param roilist: List of ROI names (keys of roidict)
        :type roilist: List
        :param dict roidict: Dict of ROI information
        :param currentroi: Name of the selected ROI or None (no selection)
        """
        if roidict is not None:
            self.setRois(roidict)
        else:
            self.setRois(roilist)
        if currentroi:
            self.setActiveRoi(currentroi)


_indexNextROI = 0


class ROI(_RegionOfInterestBase):
    """The Region Of Interest is defined by:

    - A name
    - A type. The type is the label of the x axis. This can be used to apply or
      not some ROI to a curve and do some post processing.
    - The x coordinate of the left limit (fromdata)
    - The x coordinate of the right limit (todata)

    :param str: name of the ROI
    :param fromdata: left limit of the roi
    :param todata: right limit of the roi
    :param type: type of the ROI
    """

    sigChanged = qt.Signal()
    """Signal emitted when the ROI is edited"""

    def __init__(self, name, fromdata=None, todata=None, type_=None):
        _RegionOfInterestBase.__init__(self)
        self.setName(name)
        global _indexNextROI
        self._id = _indexNextROI
        _indexNextROI += 1

        self._fromdata = fromdata
        self._todata = todata
        self._type = type_ or 'Default'

        self.sigItemChanged.connect(self.__itemChanged)

    def __itemChanged(self, event):
        """Handle name change"""
        if event == items.ItemChangedType.NAME:
            self.sigChanged.emit()

    def getID(self):
        """

        :return int: the unique ID of the ROI
        """
        return self._id

    def setType(self, type_):
        """

        :param str type_:
        """
        if self._type != type_:
            self._type = type_
            self.sigChanged.emit()

    def getType(self):
        """

        :return str: the type of the ROI.
        """
        return self._type

    def setFrom(self, frm):
        """

        :param frm: set x coordinate of the left limit
        """
        if self._fromdata != frm:
            self._fromdata = frm
            self.sigChanged.emit()

    def getFrom(self):
        """

        :return: x coordinate of the left limit
        """
        return self._fromdata

    def setTo(self, to):
        """

        :param to: x coordinate of the right limit
        """
        if self._todata != to:
            self._todata = to
            self.sigChanged.emit()

    def getTo(self):
        """

        :return: x coordinate of the right limit
        """
        return self._todata

    def getMiddle(self):
        """

        :return: middle position between 'from' and 'to' values
        """
        return 0.5 * (self.getFrom() + self.getTo())

    def toDict(self):
        """

        :return: dict containing the roi parameters
        """
        ddict = {
            'type': self._type,
            'name': self.getName(),
            'from': self._fromdata,
            'to': self._todata,
        }
        if hasattr(self, '_extraInfo'):
            ddict.update(self._extraInfo)
        return ddict

    @staticmethod
    def _fromDict(dic):
        assert 'name' in dic
        roi = ROI(name=dic['name'])
        roi._extraInfo = {}
        for key in dic:
            if key == 'from':
                roi.setFrom(dic['from'])
            elif key == 'to':
                roi.setTo(dic['to'])
            elif key == 'type':
                roi.setType(dic['type'])
            else:
                roi._extraInfo[key] = dic[key]

        return roi

    def isICR(self):
        """

        :return: True if the ROI is the `ICR`
        """
        return self.getName() == 'ICR'

    def computeRawAndNetCounts(self, curve):
        """Compute the Raw and net counts in the ROI for the given curve.

        - Raw count: Points values sum of the curve in the defined Region Of
           Interest.

          .. image:: img/rawCounts.png

        - Net count: Raw counts minus background

          .. image:: img/netCounts.png

        :param CurveItem curve:
        :return tuple: rawCount, netCount
        """
        assert isinstance(curve, items.Curve) or curve is None

        if curve is None:
            return None, None

        x = curve.getXData(copy=False)
        y = curve.getYData(copy=False)

        idx = numpy.nonzero((self._fromdata <= x) &
                            (x <= self._todata))[0]
        if len(idx):
            xw = x[idx]
            yw = y[idx]
            rawCounts = yw.sum(dtype=numpy.float64)
            deltaX = xw[-1] - xw[0]
            deltaY = yw[-1] - yw[0]
            if deltaX > 0.0:
                slope = (deltaY / deltaX)
                background = yw[0] + slope * (xw - xw[0])
                netCounts = (rawCounts -
                             background.sum(dtype=numpy.float64))
            else:
                netCounts = 0.0
        else:
            rawCounts = 0.0
            netCounts = 0.0
        return rawCounts, netCounts

    def computeRawAndNetArea(self, curve):
        """Compute the Raw and net counts in the ROI for the given curve.

        - Raw area: integral of the curve between the min ROI point and the
           max ROI point to the y = 0 line.

          .. image:: img/rawArea.png

        - Net area: Raw counts minus background

          .. image:: img/netArea.png

        :param CurveItem curve:
        :return tuple: rawArea, netArea
        """
        assert isinstance(curve, items.Curve) or curve is None

        if curve is None:
            return None, None

        x = curve.getXData(copy=False)
        y = curve.getYData(copy=False)

        y = y[(x >= self._fromdata) & (x <= self._todata)]
        x = x[(x >= self._fromdata) & (x <= self._todata)]

        if x.size == 0:
            return 0.0, 0.0

        rawArea = numpy.trapz(y, x=x)
        # to speed up and avoid an intersection calculation we are taking the
        # closest index to the ROI
        closestXLeftIndex = (numpy.abs(x - self.getFrom())).argmin()
        closestXRightIndex = (numpy.abs(x - self.getTo())).argmin()
        yBackground = y[closestXLeftIndex], y[closestXRightIndex]
        background = numpy.trapz(yBackground, x=x)
        netArea = rawArea - background
        return rawArea, netArea

    @docstring(_RegionOfInterestBase)
    def contains(self, position):
        return self._fromdata <= position[0] <= self._todata


class _RoiMarkerManager(object):
    """
    Deal with all the ROI markers
    """
    def __init__(self):
        self._roiMarkerHandlers = {}
        self._middleROIMarkerFlag = False
        self._showAllMarkers = False
        self._activeRoi = None

    def setActiveRoi(self, roi):
        self._activeRoi = roi
        self.updateAllMarkers()

    def setShowAllMarkers(self, show):
        if show != self._showAllMarkers:
            self._showAllMarkers = show
            self.updateAllMarkers()

    def add(self, roi, markersHandler):
        assert isinstance(roi, ROI)
        assert isinstance(markersHandler, _RoiMarkerHandler)
        if roi.getID() in self._roiMarkerHandlers:
            raise ValueError('roi with the same ID already existing')
        else:
            self._roiMarkerHandlers[roi.getID()] = markersHandler

    def getMarkerHandler(self, roiID):
        if roiID in self._roiMarkerHandlers:
            return self._roiMarkerHandlers[roiID]
        else:
            return None

    def clear(self):
        roisHandler = list(self._roiMarkerHandlers.values())
        for roiHandler in roisHandler:
            self.remove(roiHandler.roi)

    def remove(self, roi):
        if roi is None:
            return
        assert isinstance(roi, ROI)
        if roi.getID() in self._roiMarkerHandlers:
            self._roiMarkerHandlers[roi.getID()].clear()
            del self._roiMarkerHandlers[roi.getID()]

    def hasMarker(self, markerID):
        assert type(markerID) is str
        return self.getMarker(markerID) is not None

    def changePosition(self, markerID, x):
        markerHandler = self.getMarker(markerID)
        if markerHandler is None:
            raise ValueError('Marker %s not register' % markerID)
        markerHandler.changePosition(markerID=markerID, x=x)

    def updateMarker(self, markerID):
        markerHandler = self.getMarker(markerID)
        if markerHandler is None:
            raise ValueError('Marker %s not register' % markerID)
        roiID = self.getRoiID(markerID)
        visible = (self._activeRoi and self._activeRoi.getID() == roiID) or self._showAllMarkers is True
        markerHandler.setVisible(visible)
        markerHandler.updateAllMarkers()

    def updateRoiMarkers(self, roiID):
        if roiID in self._roiMarkerHandlers:
            visible = ((self._activeRoi and self._activeRoi.getID() == roiID)
                       or self._showAllMarkers is True)
            _roi = self._roiMarkerHandlers[roiID]._roi()
            if _roi and not _roi.isICR():
                self._roiMarkerHandlers[roiID].showMiddleMarker(self._middleROIMarkerFlag)
            self._roiMarkerHandlers[roiID].setVisible(visible)
            self._roiMarkerHandlers[roiID].updateMarkers()

    def getMarker(self, markerID):
        assert type(markerID) is str
        for marker in list(self._roiMarkerHandlers.values()):
            if marker.hasMarker(markerID):
                return marker

    def updateMarkers(self):
        for markerHandler in list(self._roiMarkerHandlers.values()):
            markerHandler.updateMarkers()

    def getRoiID(self, markerID):
        for roiID, markerHandler in self._roiMarkerHandlers.items():
            if markerHandler.hasMarker(markerID):
                return roiID
        return None

    def setShowMiddleMarkers(self, show):
        self._middleROIMarkerFlag = show
        self._roiMarkerHandlers.updateAllMarkers()

    def updateAllMarkers(self):
        for roiID in self._roiMarkerHandlers:
            self.updateRoiMarkers(roiID)

    def getVisibleRois(self):
        res = {}
        for roiID, roiHandler in self._roiMarkerHandlers.items():
            markers = (roiHandler.getMarker('min'), roiHandler.getMarker('max'),
                       roiHandler.getMarker('middle'))
            for marker in markers:
                if marker.isVisible():
                    if roiID not in res:
                        res[roiID] = []
                    res[roiID].append(marker)
        return res


class _RoiMarkerHandler(object):
    """Used to deal with ROI markers used in ROITable"""
    def __init__(self, roi, plot):
        assert roi and isinstance(roi, ROI)
        assert plot

        self._roi = weakref.ref(roi)
        self._plot = weakref.ref(plot)
        self._draggable = False if roi.isICR() else True
        self._color = 'black' if roi.isICR() else 'blue'
        self._displayMidMarker = False
        self._visible = True

    @property
    def draggable(self):
        return self._draggable

    @property
    def plot(self):
        return self._plot()

    def clear(self):
        if self.plot and self.roi:
            self.plot.removeMarker(self._markerID('min'))
            self.plot.removeMarker(self._markerID('max'))
            self.plot.removeMarker(self._markerID('middle'))

    @property
    def roi(self):
        return self._roi()

    def setVisible(self, visible):
        if visible != self._visible:
            self._visible = visible
            self.updateMarkers()

    def showMiddleMarker(self, visible):
        if self.draggable is False and visible is True:
            _logger.warning("ROI is not draggable. Won't display middle marker")
            return
        self._displayMidMarker = visible
        self.getMarker('middle').setVisible(self._displayMidMarker)

    def updateMarkers(self):
        if self.roi is None:
            return
        self._updateMinMarkerPos()
        self._updateMaxMarkerPos()
        self._updateMiddleMarkerPos()

    def _updateMinMarkerPos(self):
        self.getMarker('min').setPosition(x=self.roi.getFrom(), y=None)
        self.getMarker('min').setVisible(self._visible)

    def _updateMaxMarkerPos(self):
        self.getMarker('max').setPosition(x=self.roi.getTo(), y=None)
        self.getMarker('max').setVisible(self._visible)

    def _updateMiddleMarkerPos(self):
        self.getMarker('middle').setPosition(x=self.roi.getMiddle(), y=None)
        self.getMarker('middle').setVisible(self._displayMidMarker and self._visible)

    def getMarker(self, markerType):
        if self.plot is None:
            return None
        assert markerType in ('min', 'max', 'middle')
        if self.plot._getMarker(self._markerID(markerType)) is None:
            assert self.roi
            if markerType == 'min':
                val = self.roi.getFrom()
            elif markerType == 'max':
                val = self.roi.getTo()
            else:
                val = self.roi.getMiddle()

            _color = self._color
            if markerType == 'middle':
                _color = 'yellow'
            self.plot.addXMarker(val,
                                 legend=self._markerID(markerType),
                                 text=self.getMarkerName(markerType),
                                 color=_color,
                                 draggable=self.draggable)
        return self.plot._getMarker(self._markerID(markerType))

    def _markerID(self, markerType):
        assert markerType in ('min', 'max', 'middle')
        assert self.roi
        return '_'.join((str(self.roi.getID()), markerType))

    def getMarkerName(self, markerType):
        assert markerType in ('min', 'max', 'middle')
        assert self.roi
        return ' '.join((self.roi.getName(), markerType))

    def updateTexts(self):
        self.getMarker('min').setText(self.getMarkerName('min'))
        self.getMarker('max').setText(self.getMarkerName('max'))
        self.getMarker('middle').setText(self.getMarkerName('middle'))

    def changePosition(self, markerID, x):
        assert self.hasMarker(markerID)
        markerType = self._getMarkerType(markerID)
        assert markerType is not None
        if self.roi is None:
            return
        if markerType == 'min':
            self.roi.setFrom(x)
            self._updateMiddleMarkerPos()
        elif markerType == 'max':
            self.roi.setTo(x)
            self._updateMiddleMarkerPos()
        else:
            delta = x - 0.5 * (self.roi.getFrom() + self.roi.getTo())
            self.roi.setFrom(self.roi.getFrom() + delta)
            self.roi.setTo(self.roi.getTo() + delta)
            self._updateMinMarkerPos()
            self._updateMaxMarkerPos()

    def hasMarker(self, marker):
        return marker in (self._markerID('min'),
                          self._markerID('max'),
                          self._markerID('middle'))

    def _getMarkerType(self, markerID):
        if markerID.endswith('_min'):
            return 'min'
        elif markerID.endswith('_max'):
            return 'max'
        elif markerID.endswith('_middle'):
            return 'middle'
        else:
            return None


class CurvesROIDockWidget(qt.QDockWidget):
    """QDockWidget with a :class:`CurvesROIWidget` connected to a PlotWindow.

    It makes the link between the :class:`CurvesROIWidget` and the PlotWindow.

    :param parent: See :class:`QDockWidget`
    :param plot: :class:`.PlotWindow` instance on which to operate
    :param name: See :class:`QDockWidget`
    """
    sigROISignal = qt.Signal(object)
    """Deprecated signal for backward compatibility with silx < 0.7.
    Prefer connecting directly to :attr:`CurvesRoiWidget.sigRoiSignal`
    """

    def __init__(self, parent=None, plot=None, name=None):
        super(CurvesROIDockWidget, self).__init__(name, parent)

        assert plot is not None
        self.plot = plot
        self.roiWidget = CurvesROIWidget(self, name, plot=plot)
        """Main widget of type :class:`CurvesROIWidget`"""

        # convenience methods to offer a simpler API allowing to ignore
        # the details of the underlying implementation
        # (ALL DEPRECATED)
        self.calculateROIs = self.calculateRois = self.roiWidget.calculateRois
        self.setRois = self.roiWidget.setRois
        self.getRois = self.roiWidget.getRois

        self.roiWidget.sigROISignal.connect(self._forwardSigROISignal)

        self.layout().setContentsMargins(0, 0, 0, 0)
        self.setWidget(self.roiWidget)

        self.setAreaVisible = self.roiWidget.roiTable.setAreaVisible
        self.setCountsVisible = self.roiWidget.roiTable.setCountsVisible

    def _forwardSigROISignal(self, ddict):
        # emit deprecated signal for backward compatibility (silx < 0.7)
        self.sigROISignal.emit(ddict)

    def toggleViewAction(self):
        """Returns a checkable action that shows or closes this widget.

        See :class:`QMainWindow`.
        """
        action = super(CurvesROIDockWidget, self).toggleViewAction()
        action.setIcon(icons.getQIcon('plot-roi'))
        return action

    def showEvent(self, event):
        """Make sure this widget is raised when it is shown
        (when it is first created as a tab in PlotWindow or when it is shown
        again after hiding).
        """
        self.raise_()
        qt.QDockWidget.showEvent(self, event)

    @property
    def currentROI(self):
        return self.roiWidget.currentRoi



#!/usr/bin/env python
# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
Example of drop zone supporting application/x-silx-uri.

This example illustrates the support of drag&drop of silx URLs.
It provides 2 URLs (corresponding to 2 datasets) that can be dragged to
either a :class:`PlotWidget` or a QLable displaying the URL information.
"""

from __future__ import absolute_import

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "25/01/2019"

import logging
import os
import tempfile

import h5py
import numpy

import silx.io
from silx.gui import qt
from silx.gui.plot.PlotWidget import PlotWidget

_logger = logging.getLogger(__name__)
logging.basicConfig()


class DropPlotWidget(PlotWidget):
    """PlotWidget accepting drop of silx URLs"""

    def __init__(self, parent=None, backend=None):
        PlotWidget.__init__(self, parent=parent, backend=backend)
        self.setAcceptDrops(True)

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat("application/x-silx-uri"):
            event.acceptProposedAction()

    def dropEvent(self, event):
        byteString = event.mimeData().data("application/x-silx-uri")
        silxUrl = byteString.data().decode("utf-8")
        with silx.io.open(silxUrl) as h5:
            if silx.io.is_dataset(h5):
                dataset = h5[...]
            else:
                _logger.error("Unsupported URI")
                dataset = None

        if dataset is not None:
            if dataset.ndim == 1:
                self.clear()
                self.addCurve(y=dataset, x=range(dataset.size))
                event.acceptProposedAction()
            elif dataset.ndim == 2:
                self.clear()
                self.addImage(data=dataset)
                event.acceptProposedAction()
            else:
                _logger.error("Unsupported dataset")


class DropLabel(qt.QLabel):
    """Label widget accepting drop of silx URLs"""

    DEFAULT_TEXT = "Drop an URL here to display information"

    def __init__(self, parent=None):
        qt.QLabel.__init__(self, parent)
        self.setAcceptDrops(True)
        self.setUrl(silx.io.url.DataUrl())

    def setUrl(self, url):
        template = ("<html>URL information (drop an URL here to parse its information):<ul>"
                    "<li><b>file_path</b>: {file_path}</li>"
                    "<li><b>data_path</b>: {data_path}</li>"
                    "<li><b>data_slice</b>: {data_slice}</li>"
                    "<li><b>scheme</b>: {scheme}</li>"
                    "</ul></html>"
                    )

        text = template.format(
            file_path=url.file_path(),
            data_path=url.data_path(),
            data_slice=url.data_slice(),
            scheme=url.scheme())
        self.setText(text)

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat("application/x-silx-uri"):
            event.acceptProposedAction()

    def dropEvent(self, event):
        byteString = event.mimeData().data("application/x-silx-uri")
        url = silx.io.url.DataUrl(byteString.data().decode("utf-8"))
        self.setUrl(url)
        event.acceptProposedAction()


class DragLabel(qt.QLabel):
    """Label widget providing a silx URL to drag"""

    def __init__(self, parent=None, url=None):
        self._url = url
        qt.QLabel.__init__(self, parent)
        self.setText('-' if url is None else "- " + self._url.path())

    def mousePressEvent(self, event):
        if event.button() == qt.Qt.LeftButton and self._url is not None:
            mimeData = qt.QMimeData()
            mimeData.setText(self._url.path())
            mimeData.setData(
                "application/x-silx-uri",
                self._url.path().encode(encoding='utf-8'))
            drag = qt.QDrag(self)
            drag.setMimeData(mimeData)
            dropAction = drag.exec_()


class DragAndDropExample(qt.QMainWindow):
    """Main window of the example"""

    def __init__(self, parent=None, urls=()):
        super(DragAndDropExample, self).__init__(parent)
        centralWidget = qt.QWidget(self)
        layout = qt.QVBoxLayout()
        centralWidget.setLayout(layout)
        layout.addWidget(qt.QLabel(
            "Drag and drop one of the following URLs on the plot or on the URL information zone:",
            self))
        for url in urls:
            layout.addWidget(DragLabel(parent=self, url=url))

        layout.addWidget(DropPlotWidget(parent=self))
        layout.addWidget(DropLabel(parent=self))

        self.setCentralWidget(centralWidget)


def main():
    app = qt.QApplication([])
    with tempfile.TemporaryDirectory() as tempdir:
        # Create temporary file with datasets
        filename = os.path.join(tempdir, "file.h5")
        with h5py.File(filename, "w") as f:
            f['image'] = numpy.arange(10000.).reshape(100, 100)
            f['curve'] = numpy.sin(numpy.linspace(0, 2*numpy.pi, 1000))

        # Create widgets
        example = DragAndDropExample(urls=(
            silx.io.url.DataUrl(file_path=filename, data_path='/image', scheme="silx"),
            silx.io.url.DataUrl(file_path=filename, data_path='/curve', scheme="silx")))
        example.setWindowTitle("Drag&Drop URLs sample code")
        example.show()
        app.exec_()


if __name__ == "__main__":
    main()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2004-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
:mod:`silx.gui.plot.actions.histogram` provides actions relative to histograms
for :class:`.PlotWidget`.

The following QAction are available:

- :class:`PixelIntensitiesHistoAction`
"""

from __future__ import division

__authors__ = ["V.A. Sole", "T. Vincent", "P. Knobel"]
__date__ = "10/10/2018"
__license__ = "MIT"

import numpy
import logging
import weakref

from .PlotToolAction import PlotToolAction
from silx.math.histogram import Histogramnd
from silx.math.combo import min_max
from silx.gui import qt
from silx.gui.plot import items

_logger = logging.getLogger(__name__)


class _LastActiveItem(qt.QObject):

    sigActiveItemChanged = qt.Signal(object, object)
    """Emitted when the active plot item have changed"""

    def __init__(self, parent, plot):
        assert plot is not None
        super(_LastActiveItem, self).__init__(parent=parent)
        self.__plot = weakref.ref(plot)
        self.__item = None
        item = self.__findActiveItem()
        self.setActiveItem(item)
        plot.sigActiveImageChanged.connect(self._activeImageChanged)
        plot.sigActiveScatterChanged.connect(self._activeScatterChanged)

    def getPlotWidget(self):
        return self.__plot()

    def __findActiveItem(self):
        plot = self.getPlotWidget()
        image = plot.getActiveImage()
        if image is not None:
            return image
        scatter = plot.getActiveScatter()
        if scatter is not None:
            return scatter

    def getActiveItem(self):
        if self.__item is None:
            return None
        item = self.__item()
        if item is None:
            self.__item = None
        return item

    def setActiveItem(self, item):
        previous = self.getActiveItem()
        if previous is item:
            return
        if item is None:
            self.__item = None
        else:
            self.__item = weakref.ref(item)
        self.sigActiveItemChanged.emit(previous, item)

    def _activeImageChanged(self, previous, current):
        """Handle active image change"""
        plot = self.getPlotWidget()
        item = plot.getImage(current)
        if item is None:
            self.setActiveItem(None)
        elif isinstance(item, items.ImageBase):
            self.setActiveItem(item)
        else:
            # Do not touch anything, which is consistent with silx v0.12 behavior
            pass

    def _activeScatterChanged(self, previous, current):
        """Handle active scatter change"""
        plot = self.getPlotWidget()
        item = plot.getScatter(current)
        self.setActiveItem(item)


class PixelIntensitiesHistoAction(PlotToolAction):
    """QAction to plot the pixels intensities diagram

    :param plot: :class:`.PlotWidget` instance on which to operate
    :param parent: See :class:`QAction`
    """

    def __init__(self, plot, parent=None):
        PlotToolAction.__init__(self,
                                plot,
                                icon='pixel-intensities',
                                text='pixels intensity',
                                tooltip='Compute image intensity distribution',
                                parent=parent)
        self._lastItemFilter = _LastActiveItem(self, plot)
        self._histo = None
        self._item = None

    def _connectPlot(self, window):
        self._lastItemFilter.sigActiveItemChanged.connect(self._activeItemChanged)
        item = self._lastItemFilter.getActiveItem()
        self._setSelectedItem(item)
        PlotToolAction._connectPlot(self, window)

    def _disconnectPlot(self, window):
        self._lastItemFilter.sigActiveItemChanged.disconnect(self._activeItemChanged)
        PlotToolAction._disconnectPlot(self, window)
        self._setSelectedItem(None)

    def _getSelectedItem(self):
        item = self._item
        if item is None:
            return None
        else:
            return item()

    def _activeItemChanged(self, previous, current):
        if self._isWindowInUse():
            self._setSelectedItem(current)

    def _setSelectedItem(self, item):
        if item is not None:
            if not isinstance(item, (items.ImageBase, items.Scatter)):
                # Filter out other things
                return

        old = self._getSelectedItem()
        if item is old:
            return
        if old is not None:
            old.sigItemChanged.disconnect(self._itemUpdated)
        if item is None:
            self._item = None
        else:
            self._item = weakref.ref(item)
            item.sigItemChanged.connect(self._itemUpdated)
        self.computeIntensityDistribution()

    def _itemUpdated(self, event):
        if event == items.ItemChangedType.DATA:
            self.computeIntensityDistribution()

    def _cleanUp(self):
        plot = self.getHistogramPlotWidget()
        try:
            plot.remove('pixel intensity', kind='item')
        except Exception:
            pass

    def computeIntensityDistribution(self):
        """Get the active image and compute the image intensity distribution
        """
        item = self._getSelectedItem()

        if item is None:
            self._cleanUp()
            return

        if isinstance(item, items.ImageBase):
            array = item.getData(copy=False)
            if array.ndim == 3:  # RGB(A) images
                _logger.info('Converting current image from RGB(A) to grayscale\
                    in order to compute the intensity distribution')
                array = (array[:, :, 0] * 0.299 +
                         array[:, :, 1] * 0.587 +
                         array[:, :, 2] * 0.114)
        elif isinstance(item, items.Scatter):
            array = item.getValueData(copy=False)
        else:
            assert(False)

        if array.size == 0:
            self._cleanUp()
            return

        xmin, xmax = min_max(array, min_positive=False, finite=True)
        nbins = min(1024, int(numpy.sqrt(array.size)))
        data_range = xmin, xmax

        # bad hack: get 256 bins in the case we have a B&W
        if numpy.issubdtype(array.dtype, numpy.integer):
            if nbins > xmax - xmin:
                nbins = xmax - xmin

        nbins = max(2, nbins)

        data = array.ravel().astype(numpy.float32)
        histogram = Histogramnd(data, n_bins=nbins, histo_range=data_range)
        assert len(histogram.edges) == 1
        self._histo = histogram.histo
        edges = histogram.edges[0]
        plot = self.getHistogramPlotWidget()
        plot.addHistogram(histogram=self._histo,
                          edges=edges,
                          legend='pixel intensity',
                          fill=True,
                          color='#66aad7')
        plot.resetZoom()

    def getHistogramPlotWidget(self):
        """Create the plot histogram if needed, otherwise create it

        :return: the PlotWidget showing the histogram of the pixel intensities
        """
        return self._getToolWindow()

    def _createToolWindow(self):
        from silx.gui.plot.PlotWindow import Plot1D
        window = Plot1D(parent=self.plot)
        window.setWindowFlags(qt.Qt.Window)
        window.setWindowTitle('Image Intensity Histogram')
        window.setDataMargins(0.1, 0.1, 0.1, 0.1)
        window.getXAxis().setLabel("Value")
        window.getYAxis().setLabel("Count")
        return window

    def getHistogram(self):
        """Return the last computed histogram

        :return: the histogram displayed in the HistogramPlotWiget
        """
        return self._histo
